#!/usr/bin/python

# Copyright 2016 Emutex Ltd. All rights reserved.

import os
import re
import sqlite3
import sys

class Column(object):

    def __init__(self, name):
        self.name = name
        self.rows = []

    def __repr__(self):
        return ("Column: name: %s, " % self.name +
            "rows: %s" % str(self.rows))

    def add_row(self, val):
        self.rows.append(val)


class Migrator(object):

    def __init__(self, conn, schema, table, migrate_data=True):
        self.table = table
        self.db = conn
        self.cur = conn.cursor()
        self.__schema(schema)
        self.__newtable_check()
        self.__migrate_data = migrate_data
        if self.table_exists and self.__migrate_data:
            self.__newcolumns()
            self.__data()

    def __schema(self, schema):
        with open(schema, 'r') as sqlfile:
            sql = sqlfile.read()
            sqlfile.close()
            matcher = re.compile(r'CREATE TABLE %s\(.*?\)\;' % self.table, re.DOTALL)
            match = matcher.search(sql)
            if match is None:
                raise Exception("Can't match the %s table in sql" % self.table)

            self.schema = match.group(0)

            self.index_schema_list = []
            matcher = re.compile(r'CREATE INDEX \w+ ON %s\(.*?\)\;' % self.table, re.DOTALL)
            matches = matcher.finditer(sql)
            for match in matches:
                self.index_schema_list.append(match.group(0))

    def __newtable_check(self):
        print "Checking if table '%s' already exists" % (self.table)
        self.cur.execute("SELECT COUNT(*) FROM sqlite_master WHERE type='table' and name='%s'" % self.table)
        count = self.cur.fetchone()
        self.table_exists = count[0] == 1

    def __newcolumns(self):
        newcolumns = [c.strip().split(' ')[0] for c in self.schema.split('\n')]
        for c in newcolumns:
            if c in ('CREATE', 'PRIMARY', 'FOREIGN'):
                newcolumns.remove(c)

        self.newcolums = newcolumns


    def __data(self):
        self.cur.execute('SELECT * FROM %s' % self.table)
        columns = []
        for desc in self.cur.description:
            columns.append(Column(desc[0]))

        rows = self.cur.fetchall()
        for row in rows:
            i = 0
            for value in row:
                columns[i].add_row(value)
                i += 1

        self.rows = len(rows)
        self.columns = columns

    def __find(self, name):
        for column in self.columns:
            if column.name == name:
                return column

    def __check(self):
        assert len(self.newcolums) == len(self.columns)
        for current, new in zip(self.columns, self.newcolums):
            assert current.name == new
            assert len(current.rows) == self.rows

    def __create_temp(self):
        self.temptable = self.table + '_temp'
        tempschema = self.schema.replace(self.table, self.temptable, 1)
        self.cur.execute(tempschema)

    def __create_new(self):
        print("Adding new table '%s'" % (self.table))
        self.cur.execute(self.schema)

    def __migrate(self):
        values = ','.join('?' * len(self.newcolums))
        cols = ','.join(self.newcolums)
        to_insert = []

        for i in range(self.rows):
            row = []
            for col in self.columns:
                row.append(col.rows[i])
            to_insert.append(row)
        sql = "INSERT INTO {table} ({cols}) VALUES ({values})".\
            format(table=self.temptable, cols=cols, values=values)

        self.cur.executemany(sql, to_insert)

    def __create_indices(self):
        for index_schema in self.index_schema_list:
            try:
                self.cur.execute(index_schema)
            except sqlite3.OperationalError as e:
                # Index may already exist, suppress the error
                print(str(e))

    def __cleanup(self):
        delete = "DROP TABLE %s" % self.table
        self.cur.execute(delete)

        rename = "ALTER TABLE %s RENAME TO %s" % (self.temptable, self.table)
        self.cur.execute(rename)

    def add(self, name, defvalue):
        if not self.__migrate_data:
            raise Exception("Cannot add columns without migrating data")

        if self.__find(name) is not None:
            print("Column '%s' is already part of table '%s'" % (name, self.table))
            return self

        column = Column(name)
        for _ in range(self.rows):
            column.add_row(defvalue)

        print("Adding column '%s' to table '%s'" % (name, self.table))
        self.columns.append(column)
        return self

    def remove(self, name):
        if not self.__migrate_data:
            raise Exception("Cannot remove columns without migrating data")

        column = self.__find(name)
        if column is None:
            print("Column '%s' is not part of table '%s'" % (name, self.table))
            return self

        print("Removing column '%s' from table '%s'" % (name, self.table))
        self.columns.remove(column)
        return self

    def rename(self, old, new):
        if not self.__migrate_data:
            raise Exception("Cannot rename columns without migrating data")

        column = self.__find(old)
        if column is None:
            print("Column '%s' is not part of table '%s'" % (old, self.table))
            return self

        print("Renaming column '%s' to '%s'" % (old, new))
        column.name = new
        return self

    def store(self):
        if self.table_exists:
            if self.__migrate_data:
                self.__check()
                self.__create_temp()
                self.__migrate()
                self.__cleanup()
        else:
            self.__create_new()
        self.__create_indices()
        self.db.commit()

# ----------------------------- MAIN -------------------------------------------

'''
Each database change should be tracked in this script too.
Add a comment to know from which version this change makes sense,
just to have a rough idea.

Since SQLite does not support renaming or removing columns from tables,
updating a table is a bit of a pain...
This app makes all those changes while keeping the data intact. It does as
follows:

    1. Creates an in-memory representation of the columns and all their values
    using the schema found on the system.

    2. Applies all the modifications to that "table", adding new columns,
    removing and renaming old ones, etc...

    3. Creates a temporary table using the new schema specified in the schema
    file and inserts old data into new temp table. At this point there's also
    checks in place to make sure the old schema matches the new one after
    applying the changes listed below.

    4. Removes the old table and rename the temporary one.

NOTE that new columns must be appended AT THE END of the existing table
'''

if len(sys.argv) != 3:
    print("Usage: %s db_path schema_path" % sys.argv[0])
    sys.exit(1)

dbpath = sys.argv[1]
schema = sys.argv[2]

if not os.path.isfile(dbpath):
    print "Can't open '%s'" % dbpath
    sys.exit(1)

conn = sqlite3.connect(dbpath)

try:
    Migrator(conn, schema, 'sensor'). \
        remove('initValue'). \
        remove('lastValue'). \
        remove('filters'). \
        remove('publish'). \
        remove('offline'). \
        add('node_ID', None). \
        store()

    Migrator(conn, schema, 'device'). \
        remove('thisChild'). \
        store()

    Migrator(conn, schema, 'node'). \
        store()

    # For the samples table, we're not changing the table schema so
    # we can skip schema update and data migration on this potentially
    # large table, but we do want newly added indices to be created
    Migrator(conn, schema, 'samples', False). \
        store()

except:
    print "Database update failed! Reverting changes ..."
    conn.rollback()
    conn.close()
    sys.exit(1)

print "Database update complete."
conn.close()
