-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local query = require 'db/query'
local config = require 'utils/config'
local tags = require 'utils/tags'
local systags = require 'utils/systags'

local params = { ... }
local vendorID = params[1]
local deviceID = params[2]
local mqttClient = tonumber(params[3])
local ANNOUNCE_TIMER = tags.getAnnounceClientTimer ()
local SYSTAGS_TIMER = tags.getSystagsTimer (vendorID, deviceID)
local CLIENT_RECONNECT = tags.getClientReconnect ()
local CAFILE = getSystemConfigValue('cafile', '/usr/share/ubiworx/certs/ca-crt.pem')
local CERTFILE = getSystemConfigValue('certfile', '/usr/share/ubiworx/certs/client-crt.pem')
local KEYFILE = getSystemConfigValue('keyfile', '/usr/share/ubiworx/certs/client-key.pem')

local function clientStart (event)
    local activeDevices = query.
        SELECT():
        FROM('device'):
        WHERE('active', '=', 1):
        execute()

    if activeDevices.count == 0 then
        WARN ('No active devices were found')
        return
    end

    for _, row in ipairs (activeDevices.data) do
        local threadRunFlag = tags.getClientThreadRunFlag (row.vendor_ID, row.device_ID)
        local running = tagGetNum (threadRunFlag)

        if running == 0 then
            thread ('client/client', mqttClient, row.vendor_ID, row.device_ID)
        end
    end
end

local function connect (host, port, opts)
    if not host or not port then
        ERROR ('Cannot connect to broker %s:%s', tostring(host), tostring(port))
        return
    end

    if opts then
        opts = paramDecode (opts, config.INTERNAL_PARSER)
    end

    if not opts or config.toBool(opts.tls) then
        local tlsEnabled = config.isTlsEnabled(vendorID, deviceID)
        if tlsEnabled then
            if mqttTlsSet (mqttClient, CAFILE, nil, CERTFILE, KEYFILE) < 0 then
                FATAL ('Failed to set TLS for mosquitto')
            end
        end
    end

    local password = config.getMqttPassword(vendorID, deviceID)
    if mqttUsernamePwSet(mqttClient, vendorID, password) < 0 then
        FATAL ('Failed to set mosquitto credentials')
    end

    while true do
        local mqttKeepalive = 30

        local ret = mqttStart(mqttClient, host, port, mqttKeepalive)

        if ret == 0 then
            break
        end
        NOTIFY ('Client announce MQTT connect failure, sleeping ...')
        msleep (10000)
    end
end

local function disconnect ()
    mqttStop (mqttClient)
end

local function reconnect (tag, value)
    local paramTable = paramDecode (value, config.INTERNAL_PARSER)
    paramTable.port = tonumber (paramTable.port)
    if not paramTable.host or not paramTable.port then
        ERRORF ('Cannot reconnect broker, invalid parameters %q', value)
    else
        NOTIFYF ('Reconnecting to broker %s:%s', paramTable.host, paramTable.port)
        disconnect ()
        connect (paramTable.host, paramTable.port, paramTable.options)
    end
    eventReply (tag)
end

connect (config.getClientBroker(vendorID, deviceID))
clientStart ()

-- Start a 10 second timer to rescan all clients and start any threads
-- that are not running
eventHandler (CLIENT_RECONNECT, function () end)
eventHandler (ANNOUNCE_TIMER,clientStart)
timerPeriodic (ANNOUNCE_TIMER, 10000)
eventHandler (SYSTAGS_TIMER, systags.update)
timerPeriodic (SYSTAGS_TIMER, systags.updateInterval)

thread ('client/alarm', mqttClient, vendorID, deviceID)

-- Start apps
local apps = config.getValue(vendorID, deviceID, "apps", "")
local appsTable = paramDecode(apps, config.INTERNAL_PARSER)
if not appsTable then
    WARNF ('Unparseable apps config %q', apps)
else
    for _, app in pairs (appsTable) do
        NOTIFYF ('Starting thread for "%s" app', app)
        thread (string.format('apps/%s/%s', app, app), mqttClient, vendorID, deviceID)
    end
end

while true do
    local event, param = eventWait ()
    if event == CLIENT_RECONNECT then
        reconnect (event, param)
    else
        eventDispatch (event, param)
    end
end
