-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local query = require "db/query"
local sensor = require "client/sensor"
local node = require "client/node"
local config = require "utils/config"
local utils = require "utils/utils"
local usensor = require "utils/sensor"
local store_fwd = require 'client/store_forward'
local protocol = require "utils/protocol"
local errors = require "utils/errors"
local tags = require 'utils/tags'
local osPackages = require 'client/os_packages'
local dkjson = require "third_party/dkjson"
local systags = require 'utils/systags'

local params = { ... }
local mqttClient = tonumber(params[1])
local vendorID = params[2]
local deviceID = params[3]

local CLIENT_THREAD = tags.getClientThread (vendorID, deviceID)
local THREAD_RUNNING_FLAG = tags.getClientThreadRunFlag (vendorID, deviceID)
local CLIENT_TIMER = tags.getClientTimer (vendorID, deviceID)
local SCRIPTS_UPDATED = tags.getDeviceScriptsUpdated(vendorID, deviceID)
local HANDSHAKE_DONE = tags.getHandshakeDone (vendorID, deviceID)

local transactionsQuotaTimer = "transactionsQuotaTimer"
local transactionsQuotaTimerTimeout =
    tonumber(getSystemConfigValue('transactionsQuotaTimerTimeout', '600000'))
local sendSensors = true
local getConfig = true
local configTimeStamp = 0

local function handleRedirectFlag (paramTable)
    local msg = paramEncode (paramTable, config.INTERNAL_PARSER)
    eventRequest (tags.getClientReconnect (), msg)
end

local REGACK_FLAGS_HANDLER = {
    redirect = handleRedirectFlag
}

local function reconfigSensorsActuators()
    INFO ("Send samples reconfig event")
    eventPost (tags.getSensorsReconfig(), nil, true)
    eventPost (tags.getActuatorsReconfig(), nil, true)
end

local function sendConfigItem(name)

    local value = config.getValue(vendorID, deviceID, name)

    if value == '' then
        NOTIFYF('Config entry %q not found', name)
        return
    end

    local parserType = config.getParserType (vendorID, deviceID)
    local configTable = {
        action = 'setconfig',
        name = name,
        value = value,
    }

    NOTIFYF('Sending config %s=%s to server', name, value)
    local msg = paramEncodeHeader(configTable, parserType)
    mqttPublishLocal (mqttClient, tags.server, msg)
end

local function waitForServerMessage (action, timeout)

    -- Start a timer before waiting for a message.  This timer could be a
    -- keepalive timer or a timeout timer if we are waiting for an answer to
    -- a query.  We leave the timer running even if we get a message.  Only
    -- stop the timer when we are returning from the function.
    DEBUG ("CLIENT wait timer start " .. CLIENT_TIMER)
    timerStart (CLIENT_TIMER, timeout)

    while true do
        local event, params = eventWait ()

        if event == CLIENT_TIMER then
            -- No need to stop the timer, it has expired
            return nil
        end

        local paramTable = paramDecodeHeader (params)
        local parserType = config.getParserType (vendorID, deviceID)

        local command = paramTable.action


        if event == CLIENT_THREAD then
            if command == "setreply" then
                mqttPublishLocal (mqttClient, tags.server, params)
            elseif command == "getreply" then
                mqttPublishLocal (mqttClient, tags.server, params)
            elseif command == "data" then
                local topic = paramTable.topic
                -- strip action and topic off
                paramTable.topic = nil
                paramTable.action = nil
                local msg = paramEncodeHeader (paramTable, parserType)
                mqttPublish (mqttClient, topic, msg)
            else
                WARN ("WARNING client thread got unknown command " .. params)
            end

        elseif event == tags.client then
            if command == 'set' then
                local sensorID = paramTable.sensor_ID
                local value = paramTable.value
                tagSetString (sensorID, value)

                local replyTable = {
                    action = 'setreply',
                    device_ID = deviceID,
                    vendor_ID = vendorID,
                    sensor_ID = sensorID,
                    value = value
                }
                local msg = paramEncodeHeader (replyTable, parserType)
                mqttPublishLocal(mqttClient, tags.server, msg)

            elseif command == 'get' then
                local sensorID = paramTable.sensor_ID or ""
                local value = tagGetString (sensorID) or ""

                local replyTable = {
                    action = 'getreply',
                    device_ID = deviceID,
                    vendor_ID = vendorID,
                    sensor_ID = sensorID,
                    value = value
                }
                local msg = paramEncodeHeader (replyTable, parserType)
                mqttPublishLocal(mqttClient, tags.server, msg)

            elseif command == 'sendconfig' then
                sendConfigItem(paramTable.name)

            elseif command == 'newconfig' then
                -- We got an unsolicited newconfig command.  Flag that we must
                -- recheck sensors and config but continue looping for the
                -- event we are looking for
                sendSensors = true
                getConfig = true

                -- If action is nil then this is just a timer so go ahead and
                -- return to start the numeration process again
                if (action == nil) then
                    timerStop (CLIENT_TIMER)
                    return nil
                end

            elseif command == 'ackdata' then
                store_fwd.processServerNewAckdata(vendorID, deviceID, paramTable)

            elseif command == 'quota_reached' then
                NOTIFY('Got quota_reached. Deactivating all sensors')
                query.UPDATE('sensor'):
                    SET({ active = 0 }):
                    WHERE('vendor_ID', '=', vendorID):
                    AND('device_ID', '=', deviceID):
                execute()
                reconfigSensorsActuators()
                timerStart(transactionsQuotaTimer, transactionsQuotaTimerTimeout)
                return nil

            elseif (action == "" or action == nil) and (command ~= "regack") then
                -- We got a result and the caller doesn't care so we are
                -- finished.
                timerStop (CLIENT_TIMER)
                return params

            elseif command == action then
                -- Got what we wanted
                timerStop (CLIENT_TIMER)
                return params

            elseif command == "regack" then
                WARN ("Received unexpected 'regack'," ..
                    "check if you have another instance of this client running.")

            else
                -- We got a result but it is the wrong one.  Possible protocol
                -- bug.  Print a warning but keep looping - should time out.
                WARN ("WARNING unexpected action " .. command ..
                    " while looking for " .. action)
                -- return nil
            end

        elseif event == transactionsQuotaTimer then
            NOTIFY(event .. ' timer has expired. Checking for quota status')

            timerStop (CLIENT_TIMER)
            local msg = paramEncodeHeader ({action = 'quota_status'}, parserType)
            mqttPublishLocal (mqttClient, tags.server, msg)

            local serverMqttResponse = waitForServerMessage('quota', 5000)

            if serverMqttResponse == nil then
                timerStart(transactionsQuotaTimer, transactionsQuotaTimerTimeout)
                return nil
            end

            local serverResponse = paramDecodeHeader (serverMqttResponse)

            if serverResponse.status == 'OK' then
                NOTIFY('Transactions quota has been reset. Activating sensors')
                local sensors = utils.split(serverResponse.sensors, ';')

                for i = 1, #sensors do
                    NOTIFY('Activating sensor ' .. sensors[i])
                    query.UPDATE('sensor'):
                        SET({ active = 1 }):
                        WHERE('vendor_ID', '=', vendorID):
                        AND('device_ID', '=', deviceID):
                        AND('sensor_ID', '=', sensors[i]):
                        execute()
                end
                reconfigSensorsActuators()
                timerStop(transactionsQuotaTimer)
            else
                timerStart(transactionsQuotaTimer, transactionsQuotaTimerTimeout)
            end
            return nil

        else
            WARN ("WARNING unknown event " .. event)
        end

    end

    WARN ("Isn't this unreachable?")
    return nil
end

--! @brief Updates the device record in the database with
--!        the parameters received in the regack.
--! @param params parameters that came with the regack.
local function readServerResponse(params)
    params.mobile = tonumber(params.mobile)
    params.keepalive = tonumber(params.keepalive)
    params.latitude = tonumber(params.latitude)
    params.longitude = tonumber(params.longitude)
    params.restart = tonumber(params.restart)
    params.config_timestamp = tonumber(params.config_timestamp)

    if params.config_timestamp then
        configTimeStamp = params.config_timestamp
    end

    if params.restart == 1 then
        sendSensors = true;
        getConfig = true;
    end

    local valuesToSet = {}
    if params.mac then
        valuesToSet.MAC = params.mac
    end
    if params.name then
        valuesToSet.name = params.name
    end
    if params.mobile then
        valuesToSet.mobile = params.mobile
    end
    if params.keepalive then
        valuesToSet.keepalive = params.keepalive
    end
    if params.longitude then
        valuesToSet.longitude = params.longitude
    end
    if params.latitude then
        valuesToSet.latitude = params.latitude
    end

    if next(valuesToSet) ~= nil then
        query.UPDATE("device"):
            SET(valuesToSet):
            WHERE("vendor_ID", "=", vendorID):
            AND("device_ID", "=", deviceID):
            execute()
    end
    return true
end

local function register(deviceRow)
    NOTIFY(string.format("Publishing announce for %q", deviceRow.device_ID))
    local connected = true
    local registered

    local csvTable = {
        vendor_ID = deviceRow.vendor_ID,
        device_ID = deviceRow.device_ID,
        name = deviceRow.name,
        mac = deviceRow.MAC,
        latitude = deviceRow.latitude,
        longitude = deviceRow.longitude,
        keepalive = deviceRow.keepalive,
        mobile = deviceRow.mobile,
        version = protocol.VERSION
    }

    -- If the device is mobile then the system tags $LATITUDE and $LONGITUDE
    -- are dynamic and are set by the GPS.  If the system is not mobile then these
    -- tags are static are set by us using values from the device table in the
    if deviceRow.mobile == 1 then
        csvTable.latitude = tagGetNum(tags.dLatitude)
        csvTable.longitude = tagGetNum(tags.dLongitude)
    else
        tagSetNum (tags.dLatitude, csvTable.latitude)
        tagSetNum (tags.dLongitude, csvTable.longitude)
    end

    -- call systags.update to populate tags with initial values
    systags.update()

    -- Only sends device scripts metadata on first anounce
    csvTable.devScriptsMetadata = "{}"
    if tagGetNum(SCRIPTS_UPDATED) == 0 then
        NOTIFY("Sending device scripts metadata...")
        csvTable.devScriptsMetadata = dkjson.encode(usensor.getLocalScriptsMetadata())
    end

    local parserType = config.getParserType (vendorID, deviceID)
    local announceMsg = paramEncodeHeader (csvTable, parserType)
    while true do
        mqttPublish (mqttClient, tags.announce, announceMsg)
        NOTIFY (string.format("announce for %q published, waiting for ack", deviceRow.device_ID))

        local response = waitForServerMessage("regack", 10000)

        if (response ~= nil) then
            local paramTable = paramDecodeHeader (response)

            if paramTable.errorCode then
                ERROR(errors.toString(paramTable.errorCode))
                tagSetNum(HANDSHAKE_DONE, 0)
                return false
            elseif paramTable.flags then
                -- find the correct handler for the given flag
                local handler = REGACK_FLAGS_HANDLER[paramTable.flags]
                if not handler then
                    WARNF ('Received regack with unknown flag %q', response)
                else
                    handler (paramTable)
                end
            else
                if not connected then
                    WARN ("Client reconnected")
                    connected = true
                end
                registered = readServerResponse (paramTable)
                if registered and tagGetNum(HANDSHAKE_DONE) >= 0 then
                    NOTIFY ("Successfully registered")
                    tagSetNum(HANDSHAKE_DONE, 1)
                    tagSetNum(SCRIPTS_UPDATED, 1)
                end
                break
            end
        else
            tagSetNum(HANDSHAKE_DONE, 0)
            WARN ("register, no reply, looping")
            connected = false
        end
    end

    return registered
end

local function sendDeregister()
    local paramTable = {
        action = "deregister",
        vendor_ID = vendorID,
        device_ID = deviceID
    }
    local parserType = config.getParserType (vendorID, deviceID)
    local msg = paramEncodeHeader (paramTable, parserType)
    mqttPublishLocal (mqttClient, tags.server, msg)

    return waitForServerMessage("deregack", 10000)
end

local function currentConfigIsOlder()
    local resultTable = query.
        SELECT():
        FROM("device"):
        WHERE("vendor_ID", "=", vendorID):
        AND("device_ID", "=", deviceID):
        execute()
    local currentConfigTS = resultTable.data[1].config_timestamp

    return currentConfigTS < configTimeStamp
end

local function requestConfigFromServer()
    local parserType = config.getParserType (vendorID, deviceID)
    local msg = paramEncodeHeader ({action = 'getconfig'}, parserType)
    mqttPublishLocal (mqttClient, tags.server,msg)
    return waitForServerMessage("configvalue",10000)
end

local function storeReceivedConfig(msgData)
    local configTable = paramDecodeHeader (msgData)

    configTable.action = nil
    config.store(vendorID, deviceID, configTable)
    NOTIFYF ("stored configvalues in DB")

    local softwareUpgradeStatus = tagGetString(tags.softwareUpgradeStatus)
    if softwareUpgradeStatus ~= '' then
        tagSetString(tags.softwareUpgradeStatus, '')
        config.setSoftwareUpgradeStatus(vendorID, deviceID, softwareUpgradeStatus)
        config.sendValue(vendorID, deviceID, config.SOFTWARE_UPGRADE_STATUS)
    end

    osPackages.versionCheck(vendorID, deviceID)
end

local function mainClientLoop ()

    local device = query.SELECT():
        FROM('device'):
        WHERE('vendor_ID', '=', vendorID):
        AND('device_ID', '=', deviceID):
        AND('active', '=', 1):
        execute()

    local data
    if device.count < 1 then
        local msg = "Device '%s' has become inactive. Thread exiting"
        NOTIFY(msg:format(deviceID))
        return false
    end

    device = device.data[1]

    local registered = register (device)
    if not registered then
        query.UPDATE("device"):
            SET({ active = 0 }):
            WHERE("vendor_ID", "=", vendorID):
            AND("device_ID", "=", deviceID):
            execute()
        return false
    end

    if sendSensors then
        data = sendDeregister()

        if data == nil then
            return true
        end

        data = node.declareClientNodes(mqttClient, vendorID, deviceID,
            waitForServerMessage)

        if data == 'timeout' then
            return true
        end

        data = sensor.declareClientSensors(mqttClient, vendorID, deviceID,
            waitForServerMessage)

        if data == 'timeout' then
            return true
        end

        data = node.requestServerDefinedNodes (mqttClient, vendorID, deviceID,
            waitForServerMessage)

        if data == nil then
            return true
        end

        data = sensor.requestServerDefinedSensors (mqttClient, vendorID, deviceID,
            waitForServerMessage)

        if data == nil then
            return true
        end
        NOTIFYF("sync of sensors for %s.%s done", vendorID, deviceID)
        -- If sornack received means we have fully
        -- configured all our sensors.  At this point we post an
        -- event to the samples thread to reconfigure so that all
        -- active sensors are scanning and sending data.

        reconfigSensorsActuators()
        sendSensors = false
    end

    if getConfig then
        if currentConfigIsOlder() then
            data = requestConfigFromServer()
            if data == nil then
                sendSensors = true
                return true
            end
            storeReceivedConfig(data)
        end

        -- Reconfigure any local rules that are active
        INFO ("Send rules reconfig event")
        eventPost (tags.getRulesReconfig(), nil, true)

        getConfig = false
    end

    DEBUG("Keepalive " .. device.keepalive * 1000)
    waitForServerMessage(nil, device.keepalive * 1000)

    query.UPDATE('device'):
        SET({
            last_connection = os.time()
        }):
        WHERE('vendor_ID', '=', vendorID):
        AND('device_ID', '=', deviceID):
        execute();

    return true
end

if tagGetBool(THREAD_RUNNING_FLAG) then
    INFO ("client thread already running")
    return
end

tagSetBool (THREAD_RUNNING_FLAG, true)
tagSetNum (HANDSHAKE_DONE, 0)
tagSetNum  (SCRIPTS_UPDATED, 0)

eventHandler (CLIENT_THREAD , function() end)
eventHandler (CLIENT_TIMER, function() end)

mqttSubscribeLocal (mqttClient, tags.client)
eventHandler(tags.client, function() end)

-- Give some time to the mqtt topic subscription to go through
msleep(200)

while mainClientLoop() do
    -- looping
end

eventHandlerClear(tags.client)
mqttUnsubscribeLocal (mqttClient, tags.client)
timerStop(tags.getSendAllSamplesTimer(vendorID, deviceID))
tagSetNum (HANDSHAKE_DONE, 0)
tagSetBool (THREAD_RUNNING_FLAG, false)

NOTIFYF("Client thread %d.%s exiting now", vendorID, deviceID)
