-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local json = require 'third_party/dkjson'
local tags = require 'utils/tags'

local params = { ... }
local vendorID = params[1]
local deviceID = params[2]

local function getFilelist(pattern)
    pattern = '/var/log/' .. pattern .. '*'

    -- ls -tr lists the files reverse-ordered by mtime to get oldest-first
    local cmd = {
        '/bin/ls',
        '-tr', pattern,
    }
    local output = io.popen(table.concat(cmd, ' '))

    local filelist = {}
    for line in output:lines() do
        table.insert(filelist, line)
    end
    output:close()

    return table.concat(filelist, ' ')
end

local function buildCommand(files, tempfile, startline, linecount)
    -- Using -f with zcat forces it to handle both compressed and uncompressed
    -- input files without complaining (archived logs may be gzipped)
    local zcat_cmd = table.concat(
        {
            '/bin/zcat',
            '-f', files,
            '>', tempfile,
        }, ' ')

    -- Count the number of lines in the file, this will be the first line of output
    local wc_cmd = table.concat(
        {
            '/usr/bin/wc',
            '-l',
            '<', tempfile,
        }, ' ')

    -- Extract the relevant lines from the output, using either 'tail' or 'sed'
    local trim_cmd
    if linecount < 0 then
        -- If linecount is negative, get last 'linecount' lines and ignore startline
        trim_cmd = table.concat(
            {
                '/usr/bin/tail',
                '-n', math.abs(linecount),
                tempfile
            }, ' ')
    elseif linecount > 0 then
        trim_cmd = table.concat(
            {
                '/bin/sed',
                '-n', string.format("'%d,%dp'", startline, startline + linecount -1),
                tempfile
            }, ' ')
    else
        trim_cmd = table.concat(
            {
                '/bin/sed',
                '-n', string.format("'%d,$p'", startline),
                tempfile
            }, ' ')
    end

    return table.concat({zcat_cmd, wc_cmd, trim_cmd}, ' && ')
end

local function processLogRequest(event, params)
    if not params or params == "" then
        return
    end

    local options = json.decode(params)
    local startline = tonumber(options.startline) or 1
    local linecount = tonumber(options.linecount) or 100

    if not options.logfile then
        ERROR("processLogRequest: No logfile specified\n")
        tagSetString(tags.logResponse, json.encode({error = "No logfile specified"}));
        return
    end

    -- Restrict to files and sub-directories under /var/log
    local pattern = string.match(options.logfile, "^([-_A-Za-z0-9]+[/]*[-_A-Za-z0-9.]*)$")
    if not pattern or pattern == "" then
        ERRORF("processLogRequest: Invalid logfile specified: %s\n", options.logfile)
        tagSetString(tags.logResponse, json.encode({error = "Invalid logfile specified"}));
        return
    end

    local files = getFilelist(pattern)
    if not files or files == "" then
        ERRORF("processLogRequest: No logfiles found matching: %s\n", options.logfile)
        tagSetString(tags.logResponse, json.encode({error = "No matching logfiles found"}));
        return
    end
    INFO("processLogRequest: Fetching log files from " .. pattern)

    -- Use a tempfile rather than pipes between commands because the latter
    -- triggers 'broken pipe' errors when the commands are executed from lua
    local tempfile = os.tmpname()

    -- Include the request parameters in the response, as context information
    local result = options

    local cmd = buildCommand(files, tempfile, startline, linecount)
    local output = io.popen (cmd, "r")

    result.totallines = tonumber(output:read("*l"))
    if not result.totallines then
        ERROR("processLogRequest: Failed to access specified logfile\n")
        tagSetString(tags.logResponse, json.encode({error = "Failed to access specified logfile"}));
        return
    end

    -- Provide absolute values for startline, linecount and totallines in result
    -- This enables the actual position of these lines in the overall log to be
    -- calculated and thus facilitates pagination
    if linecount < 0 then
        result.linecount = math.abs(linecount)
        result.startline = result.totallines - result.linecount + 1
        if result.startline < 0 then
            result.startline = 1
        end
    elseif linecount > 0 then
        result.linecount = linecount
        result.startline = startline
    else
        result.linecount = result.totallines
        result.startline = startline
    end
    -- Corrections for cases where actual lines available was less than what was requested
    if (result.startline <= result.totallines) then
        if (result.startline + result.linecount - 1) > result.totallines then
            result.linecount = result.totallines - result.startline + 1
        end
    else
        result.linecount = 0
    end

    result.data = {}
    for line in output:lines() do
        local entry = {}
        -- Attempt to split each line from the log file into "time|host|process|message" (typical syslog format)
        local time,host,process,message = string.match(line, '^([A-Za-z]+[ ]+[0-9]+[ ][:0-9]+)[ ]([^ ]+)[ ]([^:]+)[:][ ](.*)[\r\n]*$')
        if message and message ~= "" then
            entry = {
                time = time,
                process = process,
                message = message
            }
        else
            -- If unable to parse with the above format, just store the whole line in .message
            entry = {
                message = line
            }
        end
        table.insert(result.data, entry)
    end
    os.remove(tempfile)
    output:close()

    tagSetString(tags.logResponse, json.encode(result));
end

eventHandler (tags.logRequest, processLogRequest)
while true do
    local event, param = eventWait()
    eventDispatch(event, param)
end
