-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local query = require "db/query"
local utils = require "utils/utils"
local config = require "utils/config"
local time = require "utils/time"
local tags = require "utils/tags"
local errors = require 'utils/errors'
local unode = require "utils/node"

local self = {}

--
---  Function creates a node declared by the server.
-- @param vendor is the client of mqtt.
-- @param paramTable is the params.
function self.addServerDefinedNodeToDB (mqttClient, paramTable)
    unode.numifyNodeTable (paramTable)

    if paramTable.mobile == 0 then
        paramTable.longitude = 0
        paramTable.latitude = 0
    end

    local resultTable = query.
        SELECT ():
        FROM ('node'):
        WHERE ('vendor_ID', '=', paramTable.vendor_ID):
        AND ('device_ID', '=', paramTable.device_ID):
        AND ('node_ID', '=', paramTable.node_ID):
        execute ()

    if resultTable.count == 0 then
        query.INSERT_INTO ('node'):
            VALUE_TABLE ({
                vendor_ID   = paramTable.vendor_ID,
                device_ID   = paramTable.device_ID,
                node_ID      = paramTable.node_ID,
                name        = paramTable.name,
                type        = paramTable.type,
                hwaddr      = paramTable.hwaddr,
                hwtype      = paramTable.hwtype,
                origin      = paramTable.origin,
                timestamp   = paramTable.timestamp,
                mobile      = paramTable.mobile,
                latitude    = paramTable.latitude,
                longitude   = paramTable.longitude,
                active      = 1,
                sampleRate  = paramTable.sampleRate,
            }):
            execute ()
    else
        query.UPDATE ('node'):
            SET ({
                name        = paramTable.name,
                type        = paramTable.type,
                hwaddr      = paramTable.hwaddr,
                hwtype      = paramTable.hwtype,
                origin      = paramTable.origin,
                timestamp   = paramTable.timestamp,
                mobile      = paramTable.mobile,
                latitude    = paramTable.latitude,
                longitude   = paramTable.longitude,
                active      = 1,
                sampleRate  = paramTable.sampleRate,
            }):
            WHERE ('vendor_ID', '=', paramTable.vendor_ID):
            AND ('device_ID', '=', paramTable.device_ID):
            AND ('node_ID', '=', paramTable.node_ID):
            execute ()
    end
end

---  Function activates each node is active in the database.
-- @param mqttClient is the client of mqtt.
-- @param vendor is the vendor id
-- @param device is the device id
-- @param wait
function self.requestServerDefinedNodes (mqttClient, vendor,device, wait)
    local parserType = config.getParserType (vendor, device)
    local finished = false
    local data
    while not finished do
        local msg = paramEncodeHeader ({action = 'getnode'}, parserType)
        mqttPublishLocal (mqttClient, tags.server, msg)

        -- Supply a blank expected command because we could get
        -- either a nodnack or nodack so neitehr is an error
        data = wait ("", 10000)

        if (data == nil) then
            WARN ("WARNING node get from server timed out")
            return 'timeout'
        end

        local response = paramDecodeHeader (data)
        if response.action == 'nodack' and response.errorCode == errors.nodeRemoved then
            query.UPDATE ('node'):
                SET ({ active = 0 }):
                WHERE ('vendor_ID', '=', vendor):
                AND ('device_ID', '=', device):
                AND ('node_ID', '=', response.node_ID):
                execute ()
            WARNF('Server reported node %q has been removed', response.node_ID)
        elseif response.action == "nodack" then
            self.addServerDefinedNodeToDB (mqttClient, response)
        elseif response.action == "nodnack" then
            finished = true
            NOTIFY("action=getnode: got nodnack")
        end
    end

    return data
end

function self.declareClientNodes(mqttClient, vendorID,deviceID, wait)
    local parserType = config.getParserType (vendorID, deviceID)
    local resultTable = query.SELECT():
        FROM('node'):
        WHERE('vendor_ID', '=', vendorID):
        AND('device_ID', '=', deviceID):
        AND('origin', '=', 1):
        execute()

    for _, node in ipairs (resultTable.data) do
        local latitude = node.latitude
        local longitude = node.longitude
        if node.mobile == 1 then
            latitude = tagGetNum("latitude")
            longitude = tagGetNum("longitude")
        end

        local toSend = {
            action = 'newnode',
            vendor_ID = node.vendor_ID,
            device_ID = node.device_ID,
            node_ID = node.node_ID,
            name = node.name,
            hwaddr = node.hwaddr,
            hwtype = node.hwtype,
            origin = node.origin,
            timestamp = node.timestamp,
            type = node.type,
            sampleRate = node.sampleRate,
        }
        local msg = paramEncodeHeader (toSend, parserType)
        mqttPublishLocal (mqttClient, tags.server, msg)

        local serverResponse = wait("nodack", 10000)

        if serverResponse == nil then
            WARN ("WARNING node get from server timed out")
            return 'timeout'
        end

        local responseTable = paramDecodeHeader (serverResponse)

        if responseTable.errorCode then
            query.UPDATE('node'):
                SET({ active = 0 }):
                WHERE('vendor_ID', '=', node.vendor_ID):
                AND('device_ID', '=', node.device_ID):
                AND('node_ID', '=', node.node_ID):
                execute()
            ERROR(errors.toString(responseTable.errorCode))
        elseif responseTable.action == 'nodnack' then
            self.addServerDefinedNodeToDB(mqttClient, responseTable)
        end
    end

    return 'OK'
end

return self
