-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local deviceControl = require 'utils/device_control'
local query = require 'db/query'
local config = require 'utils/config'
local storeFwd = require 'client/store_forward'
local time = require 'utils/time'
local tags = require 'utils/tags'
local utils = require 'utils/utils'
local usensor = require "utils/sensor"
local alarm = require 'utils/alarm'

local params = { ... }
local vendorID = params[1]
local deviceID = params[2]
local mqttClient = params[3]

local sensorTimers = {}
local sensorThreads = {}
local sensorRegistered = {}
local asyncSensors = {}
local lastSampleTS = {}
local sensorRange = {}
local RECONFIG = tags.getSensorsReconfig ()
local MILLISECONDS_PER_DAY = 86400000

local function tsGreaterEqual(ts1, ts2)
    if ts1.sec > ts2.sec then
        return true
    end

    if ts1.sec < ts2.sec then
        return false
    end

    return ts1.usec >= ts2.usec
end

local function triggerSensorErrorAlarm(sensorID, nowTs, lastTs, active)
    if sensorRegistered[sensorID].sensorErrorAlarm == active then
        return
    end
    sensorRegistered[sensorID].sensorErrorAlarm = active

    local msg = paramEncode ({
        alarm_ID = alarm.buildAlarmId({"SensorError", sensorID}),
        vendor_ID = vendorID,
        device_ID = deviceID,
        sensor_ID = sensorID,
        seconds = tonumber(nowTs),
        value = 0,
        category = alarm.systemCategory,
        latitude = 0,
        longitude = 0,
        notes = string.format("No update from sensor %s since %s<br>" ..
                              "Error level: %d%%",
                              sensorID, os.date("%c", lastTs),
                              tagGetNum(sensorID .. ".errorlevel") * 100),
        name = string.format("No update from sensor %s", sensorID),
        active = active,
    }, alarm.internalParser)
    tagSetString (tags.alarm, msg)
end

local function scan (event, param)
    local sensorID = param

    if utils.isIDReservedKeyword(sensorID) then
        ERRORF('Cannot publish samples of sensor ID %q', sensorID)
        return
    end

    -- Check sensor health
    local sensor = sensorRegistered[sensorID]
    if usensor.hasSampleRate(sensor.hwtype) then
        local nowTs = os.time()
        local lastTs = lastSampleTS[sensorID].sec
        -- Set maximum interval to 5 times the sample rate (converted to seconds)
        local maxInterval = sensor.sampleRate * 0.005
        -- If sampleRate greater than one day, use a maximum interval of
        -- sampleRate plus one day
        if sensor.sampleRate > MILLISECONDS_PER_DAY then
            maxInterval = (sensor.sampleRate + MILLISECONDS_PER_DAY) / 1000
        end
        if (nowTs - lastTs) > maxInterval then
            triggerSensorErrorAlarm(sensorID, nowTs, lastTs, 1)
            WARNF("No update from sensor %s since %s", sensorID, os.date("%c", lastTs))
        end
    end

    local data = sensorGet (sensorID)

    if not data.value or data.value == '' then
        return
    end

    if tsGreaterEqual(lastSampleTS[sensorID], data) then
        -- same timestamp has already been sent, avoid dup
        return
    end

    lastSampleTS[sensorID] = {
        sec = data.sec,
        usec = data.usec,
    }
    triggerSensorErrorAlarm(sensorID, os.time(), data.sec, 0)

    local mqttParams = {
        value = data.value,
        seconds = data.sec,
        useconds = data.usec,
        longitude = 0,
        latitude = 0,
        sequence = 0
    }

    if config.useStoreAndForward(vendorID, deviceID) then
        mqttParams.sequence = storeFwd.storeSample(vendorID, deviceID,
            sensorID, data.value, data.sec, data.usec, mqttParams.longitude,
            mqttParams.latitude)

        if not mqttParams.sequence then
            return
        end
    end

    storeFwd.shouldResend(vendorID, deviceID, sensorID, mqttParams)

    if config.scadaEnabled(vendorID, deviceID) then
        local parserType = config.getParserType (vendorID, deviceID)
        local topic = tags.getSensorTag(vendorID, deviceID, sensorID)
        local scada_params = {
            topic = topic,
            value = data.value
        }
        eventPost (tags.getScadaPublish (), paramEncodeHeader (scada_params,
            parserType))
    end

    -- if sample is not a number value we are done.
    if not tonumber(data.value) then
        return
    end

    local tvSample = time.sampleTimestamp(vendorID, deviceID, data.sec, data.usec)
    local alarmTable = {
        alarm_ID = alarm.buildAlarmId({alarm.limitCategory, sensorID}),
        vendor_ID = vendorID,
        device_ID = deviceID,
        sensor_ID = sensorID,
        seconds = tvSample.sec,
        value = data.value,
        category = alarm.limitCategory,
        latitude = 0,
        longitude = 0,
        notes = string.format("Value is %d but should be [%d,%d]",
                data.value,
                tonumber(sensorRange[sensorID].min),
                tonumber(sensorRange[sensorID].max)),
        name = sensorID .. " out of range",
        active = 0,
    }

    if tonumber(data.value) < tonumber(sensorRange[sensorID].min) or
        tonumber(data.value) > tonumber(sensorRange[sensorID].max) then
        alarmTable.active = 1
    end

    local msg = paramEncode (alarmTable, alarm.internalParser)
    tagSetString (tags.alarm, msg)
end


local function asyncScan (sensorID, value)
    scan('', sensorID)
end

local function configureStoreAndForward(vendorID, deviceID, sensorID)
    if config.useStoreAndForward(vendorID, deviceID) then
        local seq =
            storeFwd.selectNextSequence(vendorID, deviceID, sensorID)
        storeFwd.setSequence(vendorID, deviceID, sensorID, seq)
    end

    storeFwd.initSensorSamples(sensorID)
end

local function startSensorTimer(sensorID, sampleRate)
    if sampleRate > 0 then
        local timerEvent = tags.getSensorTimer(sensorID)
        NOTIFYF ('Starting timer %s with sample rate %s', timerEvent ,sampleRate)
        eventHandler (timerEvent, scan)
        timerPeriodic (timerEvent, sampleRate, sensorID)
        sensorTimers[sensorID] = true
    end
end

local function startSensorThread(sensorID, hwtype, hwaddr)
    NOTIFYF ('Starting thread %s.%s.%s', vendorID, deviceID, sensorID)

    local devFile = string.format ('devices/%s', hwtype)
    local hwaddrTable = paramDecode (hwaddr, config.HWADDR_PARSER)
    if not hwaddrTable then
        WARNF ("Ignoring invalid hwaddr '%s' for '%s'", hwaddr, sensorID)
    end
    hwaddrTable = utils.toArray (hwaddrTable) or {hwaddr}

    thread (devFile, vendorID, deviceID, sensorID, unpack(hwaddrTable))
    sensorThreads[sensorID] = true
end

local function startSensors ()
    NOTIFY ('Starting sensors...')

    local queryObj = query.
        SELECT('sensor_ID, type, hwtype, hwaddr, sampleRate, scale, offset, min, max'):
        FROM('sensor'):
        WHERE('vendor_ID', '=', vendorID):
        AND('device_ID', '=', deviceID):
        AND('active', '=', 1):
        AND('sensor_ID', '<>', 'alarm')

    local count = 0
    queryObj:prepare()
    for row in query.iterator(queryObj) do
        local hwtype = usensor.legacyHwType2HwType(row.hwtype, row.type)
        row.hwtype = hwtype

        if usensor.isSensor(hwtype) then
            local sensorID = row.sensor_ID

            if not utils.isIDReservedKeyword(sensorID) then
                count = count + 1
                if sensorRegister (sensorID, row.scale, row.offset) then
                    sensorRegistered[sensorID] = row
                    sensorRegistered[sensorID].sensorErrorAlarm = 0
                end

                sensorRange[sensorID] = {
                    min = row.min,
                    max = row.max,
                }

                -- Only initialise once avoiding sending duplicate samples
                if not lastSampleTS[sensorID] then
                    lastSampleTS[sensorID] = {
                        sec = os.time() - row.sampleRate / 1000,
                        usec = 0
                    }
                end
                if usensor.hasStorage(hwtype) then
                    configureStoreAndForward(vendorID, deviceID, sensorID)
                end
                if usensor.hasSampleRate(hwtype) then
                    startSensorTimer(sensorID, row.sampleRate)
                end
                if usensor.isAsynchronous(hwtype) then
                    eventHandler (sensorID, asyncScan)
                    asyncSensors[sensorID] = true
                end
                -- Only add new sensor thread if there is no thread for this sensor already
                if usensor.hasThread(hwtype) then
                    startSensorThread(sensorID, row.hwtype, row.hwaddr)
                end
            else
                ERRORF('%q cannot be used as a sensor ID', sensorID)
            end

        end
    end
    queryObj:finish()

    storeFwd.initSamplesTimer(vendorID, deviceID)

    NOTIFYF ('Configured %d sensors', count)
end

local function stopSensors ()
    NOTIFY ('Stopping sensors...')

    for sensorID,_ in pairs(sensorThreads) do
        deviceControl.stopDeviceThread(vendorID, deviceID, sensorID)
    end

    for sensorID,_ in pairs(sensorTimers) do
        local timerEvent = tags.getSensorTimer(sensorID)
        NOTIFYF ('Stopping timer %s', timerEvent)
        timerStop(timerEvent)
        eventHandlerClear(timerEvent)
    end

    for sensorID,_ in pairs(asyncSensors) do
        eventHandlerClear(sensorID)
    end

    for sensorID,_ in pairs(sensorRegistered) do
        sensorUnregister (sensorID)
    end

    -- clear sensor lists
    sensorTimers = {}
    sensorThreads = {}
    sensorRegistered = {}
    sensorRange = {}
    storeFwd.purgeSensorsSamples(vendorID, deviceID)
    if config.useStoreAndForward(vendorID, deviceID) then
        storeFwd.resetSequences()
    end
    storeFwd.stopSamplesTimer(vendorID, deviceID)
end

local function sensorsConfig ()
    stopSensors ()
    startSensors()
end

-- do an initial config
sensorsConfig()

eventHandler (RECONFIG, function() end)

while true do
    local event, param = eventWait()
    if event == RECONFIG then
        -- sensors_config() uses eventWait() which calls a yield
        -- yielding is not possible in event handlers
        -- so config() has to be executed outside an event handler
        sensorsConfig()
    else
        eventDispatch(event, param)
    end
end
