-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local query = require "db/query"
local utils = require "utils/utils"
local config = require "utils/config"
local time = require "utils/time"
local tags = require "utils/tags"
local errors = require 'utils/errors'
local usensor = require "utils/sensor"

local self = {}

--
---  Function creates a sensor declared by the server.
-- @param vendor is the client of mqtt.
-- @param paramTable is the params.
function self.addServerDefinedSensorToDB (mqttClient, paramTable)
    usensor.numifySensorTable (paramTable)
    paramTable.units = paramTable.units or 'degrees'

    if paramTable.mobile == 0 then
        paramTable.longitude = 0
        paramTable.latitude = 0
    end

    local resultTable = query.
        SELECT ():
        FROM ('sensor'):
        WHERE ('vendor_ID', '=', paramTable.vendor_ID):
        AND ('device_ID', '=', paramTable.device_ID):
        AND ('sensor_ID', '=', paramTable.sensor_ID):
        execute ()

    if resultTable.count == 0 then
        query.INSERT_INTO ('sensor'):
            VALUE_TABLE ({
                vendor_ID   = paramTable.vendor_ID,
                device_ID   = paramTable.device_ID,
                node_ID     = paramTable.node_ID,
                sensor_ID   = paramTable.sensor_ID,
                name        = paramTable.name,
                type        = paramTable.type,
                hwaddr      = paramTable.hwaddr,
                hwtype      = paramTable.hwtype,
                scale       = paramTable.scale,
                offset      = paramTable.offset,
                max         = paramTable.max,
                min         = paramTable.min,
                origin      = paramTable.origin,
                timestamp   = paramTable.timestamp,
                units       = paramTable.units,
                mobile      = paramTable.mobile,
                latitude    = paramTable.latitude,
                longitude   = paramTable.longitude,
                active      = 1,
                sampleRate  = paramTable.sampleRate,
            }):
            execute ()
    else
        query.UPDATE ('sensor'):
            SET ({
                name        = paramTable.name,
                type        = paramTable.type,
                hwaddr      = paramTable.hwaddr,
                hwtype      = paramTable.hwtype,
                scale       = paramTable.scale,
                offset      = paramTable.offset,
                max         = paramTable.max,
                min         = paramTable.min,
                origin      = paramTable.origin,
                timestamp   = paramTable.timestamp,
                units       = paramTable.units,
                mobile      = paramTable.mobile,
                latitude    = paramTable.latitude,
                longitude   = paramTable.longitude,
                active      = 1,
                sampleRate  = paramTable.sampleRate,
            }):
            WHERE ('vendor_ID', '=', paramTable.vendor_ID):
            AND ('device_ID', '=', paramTable.device_ID):
            AND ('sensor_ID', '=', paramTable.sensor_ID):
            execute ()
    end
end

---  Function activates each sensor is active in the database.
-- @param mqttClient is the client of mqtt.
-- @param vendor is the vendor id
-- @param device is the device id
-- @param wait
function self.requestServerDefinedSensors (mqttClient, vendor,device, wait)
    local parserType = config.getParserType (vendor, device)
    local finished = false
    local data
    while not finished do
        local msg = paramEncodeHeader ({action = 'getsensor'}, parserType)
        mqttPublishLocal (mqttClient, tags.server, msg)

        -- Supply a blank expected command because we could get
        -- either a sornack or sorack so neitehr is an error
        data = wait ("", 10000)

        if (data == nil) then
            WARN ("WARNING sensor get from server timed out")
            return nil
        end

        local response = paramDecodeHeader (data)
        if response.action == 'sorack' and response.errorCode == errors.sensorRemoved then
            query.UPDATE ('sensor'):
                SET ({ active = 0 }):
                WHERE ('vendor_ID', '=', vendor):
                AND ('device_ID', '=', device):
                AND ('sensor_ID', '=', response.sensor_ID):
                execute ()
            WARNF('Server reported sensor %q has been removed', response.sensor_ID)
        elseif response.action == "sorack" then
            self.addServerDefinedSensorToDB (mqttClient, response)
        elseif response.action == "sornack" then
            finished = true
            NOTIFY("action=getsensor: got sornack")
        end
    end

    return data
end

function self.declareClientSensors(mqttClient, vendorID,deviceID, wait)
    local parserType = config.getParserType (vendorID, deviceID)
    local resultTable = query.SELECT():
        FROM('sensor'):
        WHERE('vendor_ID', '=', vendorID):
        AND('device_ID', '=', deviceID):
        AND('origin', '=', 1):
        execute()

    for _, sensor in ipairs (resultTable.data) do
        local latitude = sensor.latitude
        local longitude = sensor.longitude
        if sensor.mobile == 1 then
            latitude = tagGetNum("latitude")
            longitude = tagGetNum("longitude")
        end

        local toSend = {
            action = 'newsensor',
            vendor_ID = sensor.vendor_ID,
            device_ID = sensor.device_ID,
            node_ID = sensor.node_ID,
            sensor_ID = sensor.sensor_ID,
            name = sensor.name,
            hwaddr = sensor.hwaddr,
            hwtype = sensor.hwtype,
            units = sensor.units,
            origin = sensor.origin,
            timestamp = sensor.timestamp,
            scale = sensor.scale,
            max = sensor.max,
            min = sensor.min,
            type = sensor.type,
            offset = sensor.offset,
            sampleRate = sensor.sampleRate,
        }
        local msg = paramEncodeHeader (toSend, parserType)
        mqttPublishLocal (mqttClient, tags.server, msg)

        local serverResponse = wait("sorack", 10000)

        if serverResponse == nil then
            WARN ("WARNING sensor get from server timed out")
            return 'timeout'
        end

        local responseTable = paramDecodeHeader (serverResponse)

        if responseTable.errorCode then
            query.UPDATE('sensor'):
                SET({ active = 0 }):
                WHERE('vendor_ID', '=', sensor.vendor_ID):
                AND('device_ID', '=', sensor.device_ID):
                AND('sensor_ID', '=', sensor.sensor_ID):
                execute()
            ERROR(errors.toString(responseTable.errorCode))
        elseif responseTable.action == 'sornack' then
            self.addServerDefinedSensorToDB(mqttClient, responseTable)
        end
    end

    return 'OK'
end

return self
