-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local query = require 'db/query'
local time = require 'utils/time'
local tags = require 'utils/tags'
local config = require 'utils/config'

local self = {}
self.sequence = {}
self.state = {}
self.resendAck = {}
self.sensorSamples = {}
self.sampleTimers = {}

local state_send = 'SEND'
local state_resend = 'RESEND'

function self.setMQTTClient(mqtt)
    self.mqttClient = mqtt
end

function self.setSequence(vendorID, deviceID, sensorID, seq)
    local topic = tags.getSensorTag(vendorID, deviceID, sensorID)
    self.sequence[topic] = seq
end

function self.getSequence(vendorID, deviceID, sensorID)
    local topic = tags.getSensorTag(vendorID, deviceID, sensorID)
    return self.sequence[topic]
end

function self.resetSequences()
    self.sequence = {}
end

function self.storeSample(vendorID, deviceID, sensorID, value, sec, usec, longitude, latitude)
    local topic = tags.getSensorTag(vendorID, deviceID, sensorID)
    local seq = self.sequence[topic]
    assert(seq, string.format("got no sequence number for %s", topic))

    local ret = query.
        INSERT_INTO('samples'):
        COLUMNS('vendor_ID', 'device_ID', 'sensor_ID', 'value',
                'seconds', 'useconds', 'latitude', 'longitude', 'sequence'):
        VALUES(vendorID, deviceID, sensorID, value, sec, usec, latitude, longitude, seq):
        execute()

    if not ret then
        ERROR("Failed to store a sample for store and forward")
        return
    end

    self.sequence[topic] = self.sequence[topic] + 1
    return seq
end

function self.selectNextSequence(vendorID, deviceID, sensorID)
    local samples = query.
        SELECT('MAX(sequence) as max'):
        FROM('samples'):
        WHERE('vendor_ID', '=', vendorID):
        AND('device_ID', '=', deviceID):
        AND('sensor_ID', '=', sensorID):
        execute()

    local seq = samples.data[1].max

    if seq == nil then
        return 1
    end
    return tonumber(seq) + 1
end

function self.purgeSamples(vendorID, deviceID, sensorID, sequence)
    assert(type(sequence) == 'number')
    query.
        DELETE():
        FROM('samples'):
        WHERE('vendor_ID', '=', vendorID):
        AND('device_ID', '=', deviceID):
        AND('sensor_ID', '=', sensorID):
        AND('sequence', '<', sequence):
        execute()
end

function self.discardSample(vendorID, deviceID, sensorID, sequence)
    query.AS_TRANSACTION ({
        query.DELETE():
            FROM('samples'):
            WHERE('vendor_ID', '=', vendorID):
            AND('device_ID', '=', deviceID):
            AND('sensor_ID', '=', sensorID):
            AND('sequence', '=', sequence),

        query.UPDATE('samples'):
            SETV({ sequence = 'sequence - 1'}):
            WHERE('vendor_ID', '=', vendorID):
            AND('device_ID', '=', deviceID):
            AND('sensor_ID', '=', sensorID):
            AND('sequence', '>', sequence),
    })
end

function self.purgeAllSamples(vendorID, deviceID)
    query.
        DELETE():
        FROM('samples'):
        WHERE('vendor_ID', '=', vendorID):
        AND('device_ID', '=', deviceID):
        execute()
end

function self.getSamples(vendorID, deviceID, sensorID, sequence, maxSequencesResend)
    assert(type(sequence) == 'number')

    return query.
        SELECT('value', 'seconds', 'useconds', 'sequence', 'latitude', 'longitude'):
        FROM('samples'):
        WHERE('vendor_ID', '=', vendorID):
        AND('device_ID', '=', deviceID):
        AND('sensor_ID', '=', sensorID):
        AND('sequence', '>', sequence):
        AND('sequence', '<=', sequence + maxSequencesResend):
        execute()
end

function self.resendMissingSamples(vendorID, deviceID, sensorID, sequence)
    assert(type(sequence) == 'number')
    local topic = tags.getSensorTag(vendorID, deviceID, sensorID)
    local maxSequencesResend = config.getClientResendBatchSize(vendorID, deviceID)
    local samples = self.getSamples(vendorID, deviceID, sensorID, sequence, maxSequencesResend)

    NOTIFY('Resending ' .. samples.count .. ' samples for ' .. sensorID .. ', from seq=' .. sequence)

    if samples.count == 0 then
        local msg = 'Something went wrong. ' ..
            'There are no samples to be re-sent for sensor %s.'
        ERRORF (msg, sensorID)
        return
    end

    self.sendSamplesBatch(vendorID, deviceID, sensorID, samples.data)

    -- Selecting less samples than the maximum means there's no more to be re-
    -- sent so we can go back to normal mode and let the caller know that it
    -- can keep sending samples.
    if samples.count < maxSequencesResend then
        self.state[topic] = state_send
    else
        self.resendAck[topic] = self.resendAck[topic] + samples.count
    end

    local samplesTemp = {}
    local lastSentSeq = samples.data[samples.count].sequence
    for _, sample in pairs(self.sensorSamples[sensorID]) do
        if sample.sequence > lastSentSeq then
            table.insert(samplesTemp, sample)
        end
    end

    self.sensorSamples[sensorID] = samplesTemp
end

function self.initSensorSamples(sensorID)
    self.sensorSamples[sensorID] = {}
end

function self.sendSamplesBatch(vendorID, deviceID, sensorID, samples)
    local samplesToSend = {}
    local parserType = config.getParserType (vendorID, deviceID)
    for _, sample in pairs(samples) do
        local sampleTs = time.sampleTimestamp(vendorID, deviceID,
            sample.seconds, sample.useconds)

        sample.seconds, sample.useconds = sampleTs.sec, sampleTs.usec
        table.insert(samplesToSend, paramEncodeHeader (sample, parserType))
    end

    mqttPublishLocal (self.mqttClient, sensorID, table.concat(samplesToSend, ';'))
end

function self.sendSamples(vendorID, deviceID, sensorID)

    if tagGetNum(tags.getHandshakeDone (vendorID, deviceID)) <= 0 then
        return
    end

    NOTIFYF ("Publishing samples of sensor %q", sensorID)
    self.sendSamplesBatch(vendorID, deviceID, sensorID, self.sensorSamples[sensorID])

    self.initSensorSamples(sensorID)
end

function self.purgeSensorsSamples(vendorID, deviceID)
    for sensorID, _ in pairs(self.sensorSamples) do
        if #self.sensorSamples[sensorID] > 0 then
            local topic = tags.getSensorTag(vendorID, deviceID, sensorID)
            if self.state[topic] ~= state_resend then
                self.sendSamples(vendorID, deviceID, sensorID)
            end
        end
    end
end

function self.shouldResend(vendorID, deviceID, sensorID, mqttParams)
    local topic = tags.getSensorTag(vendorID, deviceID, sensorID)

    if config.useStoreAndForward(vendorID, deviceID) and
        self.state[topic] == state_resend then

        self.resendMissingSamples(vendorID, deviceID, sensorID, self.resendAck[topic])
        return
    end

    table.insert(self.sensorSamples[sensorID], mqttParams)

    if #self.sensorSamples[sensorID] >= config.getClientSendBatchSize(vendorID, deviceID) then
        self.sendSamples(vendorID, deviceID, sensorID)
    end
end

function self.processServerNewAckdata(vendorID, deviceID, serverMessage)
    local ack = assert(tonumber(serverMessage.ack))
    local sensorID = assert(serverMessage.sensor_ID)
    local flags = serverMessage.flags
    local topic = tags.getSensorTag(vendorID, deviceID, sensorID)

    if not self.sequence[topic] then
        return
    end

    NOTIFY(string.format('Got ackdata from the server. sensor=%s, ' ..
        'ack=%d, flags=%s', sensorID, ack, tostring(flags)))

    self.state[topic] = self.state[topic] or state_send

    if flags == nil then
        self.purgeSamples(vendorID, deviceID, sensorID, ack)
    end

    if self.state[topic] == state_send then
        if flags == 'resend' then
            self.state[topic] = state_resend
            self.resendAck[topic] = ack
        elseif flags == 'dup' and ack > self.sequence[topic] then
            self.sequence[topic] = ack + 1
        elseif flags == 'discard' then
            self.state[topic] = state_resend
            self.resendAck[topic] = ack
            self.sequence[topic] = self.sequence[topic] - 1
            self.discardSample(vendorID, deviceID, sensorID, ack + 1)
        end

    elseif self.state[topic] == state_resend and flags then -- resend or dup
        self.resendAck[topic] = ack

        if flags == 'discard' then
            self.discardSample(vendorID, deviceID, sensorID, ack + 1)
            self.sequence[topic] = self.sequence[topic] - 1
        end
    end

end

local function sendAllSamplesHandler(event, param)
    local info = paramDecode(param, config.INTERNAL_PARSER)
    NOTIFYF("Sending all samples of '%s.%s' now", info.vendorID, info.deviceID)
    self.purgeSensorsSamples(info.vendorID, info.deviceID)
end

function self.initSamplesTimer(vendorID, deviceID)
    local timer = tags.getSendAllSamplesTimer(vendorID, deviceID)

    if self.sampleTimers[timer] then
        return
    end

    eventHandler (timer, sendAllSamplesHandler)
    local info = {
        vendorID = vendorID,
        deviceID = deviceID,
    }
    timerPeriodic (timer, 60000, paramEncode(info, config.INTERNAL_PARSER))
    self.sampleTimers[timer] = true
end

function self.stopSamplesTimer(vendorID, deviceID)
    local timer = tags.getSendAllSamplesTimer(vendorID, deviceID)

    if not self.sampleTimers[timer] then
        return
    end

    timerStop(timer)
    eventHandlerClear(timer)
    self.sampleTimers[timer] = nil
end

return self
