-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

--- A class to build delete queries
-- This class makes use of the builder pattern to allow a fluent and flexible
-- query building process.
-- @classmod Delete

local checkQuery = require "db/check_query"
local Where = require "db/where"

local Delete = {}
Delete.__index = Delete

--- Constructor to create a new object of this class.
-- @treturn Delete new object.
function Delete.create()
    local deleteObj = {}
    setmetatable(deleteObj, Delete)
    return deleteObj
end

--- Set the table to delete from.
-- This defines the part of the query after the FROM keyword. This function has
-- to be called at least once to make the query valid.
-- @tparam string tableName name of the table where rows should be deleted.
-- @treturn Delete reference to the calling object.
function Delete:FROM(tableName)
    assert(type(tableName) == "string")
    self._from = tableName
    return self
end

--- Start a conditional part of the query.
-- This is optional and not required for a valid query.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Delete reference to the calling object.
function Delete:WHERE(columnName, operator, value)
    assert(not self._where)
    self._where = Where.create(columnName, operator, value)
    return self
end

--- Add a condition for the query.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Delete reference to the calling object.
function Delete:CONDITION(columnName, operator, value)
    assert(self._where)
    self._where:CONDITION(columnName, operator, value)
    return self
end

--- Concatenate two conditions with AND.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Delete reference to the calling object.
function Delete:AND(columnName, operator, value)
    assert(self._where)
    self._where:AND(columnName, operator, value)
    return self
end

--- Concatenate two conditions with OR.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Delete reference to the calling object.
function Delete:OR(columnName, operator, value)
    assert(self._where)
    self._where:OR(columnName, operator, value)
    return self
end

--- Checks if the query and its parameters are valid.
-- @treturn boolean true if the query is ok, else false.
function Delete:check()
    return (not self._where or self._where:check())
end

--- Creates a string representation of the Delete object.
-- @treturn string string representation of the query if it is valid, else nil.
function Delete:toString()
    assert(self._from)
    if not self:check() then
        return nil
    end

    local resultTable = {
        "DELETE",
        "FROM",
        self._from
    }
    if self._where then
        table.insert(resultTable, self._where:toString())
    end

    return string.format("%s;", table.concat(resultTable, " "))
end

--- Executes the query.
-- @treturn boolean true if the query was successfully executed, else false.
function Delete:execute()
    local queryStr = self:toString()
    if not queryStr then
        return false
    end

    local deletions = dbExec(queryStr)
    return deletions >= 0, deletions
end

return Delete
