-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local query = require 'db/query'
local DAO = require 'db/dao'
local Sensor = require 'db/sensor'
local Node = require 'db/node'
local usensor = require 'utils/sensor'

local Device = {}
local DeviceMetaClass = { __index = Device }
setmetatable(Device, { __index = DAO })

function Device.__deviceFromdb(vendorid, deviceid)
    local device = query.
        SELECT():
        FROM('device'):
        WHERE('vendor_ID', '=', vendorid):
        AND('device_ID', '=', deviceid):
        execute()
    assert(device.count == 1)
    return device.data[1]
end

function Device.__sensorsFromdb(vendorid, deviceid)
    local sensors = query.
        SELECT():
        FROM('sensor'):
        WHERE('vendor_ID', '=', vendorid):
        AND('device_ID', '=', deviceid):
        execute()
    return sensors.data
end

function Device.__nodesFromdb(vendorid, deviceid)
    local nodes = query.
        SELECT():
        FROM('node'):
        WHERE('vendor_ID', '=', vendorid):
        AND('device_ID', '=', deviceid):
        execute()
    return nodes.data
end

function Device.new(vendorid, deviceid)
    local device = DAO.new(Device.__deviceFromdb(vendorid, deviceid))
    local sensors = Device.__sensorsFromdb(vendorid, deviceid)
    local nodes = Device.__nodesFromdb(vendorid, deviceid)
    setmetatable(device, DeviceMetaClass)

    device._vendorid = tostring(vendorid)
    device._deviceid = tostring(deviceid)
    device._sensors = {}
    device._newsensors = {}
    for _, sensor in ipairs(sensors) do
        device._sensors[sensor.sensor_ID] = Sensor.new(sensor)
    end
    device._nodes = {}
    device._newnodes = {}
    for _, node in ipairs(nodes) do
        device._nodes[node.node_ID] = Node.new(node)
    end
    return device
end

function Device:refresh()
    self._table = Device.__deviceFromdb(self._vendorid, self._deviceid)
    self._newsensors = {}
    self._sensors = {}
    local sensors = Device.__sensorsFromdb(self._vendorid, self._deviceid)
    for _, sensor in ipairs(sensors) do
        self._sensors[sensor.sensor_ID] = Sensor.new(sensor)
    end
    self._newnodes = {}
    self._nodes = {}
    local nodes = Device.__nodesFromdb(self._vendorid, self._deviceid)
    for _, node in ipairs(nodes) do
        self._nodes[node.node_ID] = Node.new(node)
    end
    self:clearDirtyFields()
end

function Device:isLegacy()
    return usensor.getLocalScriptsMetadata().legacy
end

function Device:sensor(sensorid)
    return self._sensors[sensorid]
end

function Device:foreachSensor(cb)
    assert(type(cb) == 'function')
    for _, sensor in pairs(self._sensors) do
        cb(sensor)
    end
end

function Device:sensorNew(sensordata)
    local sensorid = assert(sensordata.sensor_ID)
    local sensor = Sensor.new(sensordata)
    self._sensors[sensorid] = sensor
    self._newsensors[sensorid] = true
    return sensor
end

function Device:node(nodeid)
    return self._nodes[nodeid]
end

function Device:foreachNode(cb)
    assert(type(cb) == 'function')
    for _, node in pairs(self._nodes) do
        cb(node)
    end
end

function Device:nodeNew(nodedata)
    local nodeid = assert(nodedata.node_ID)
    local node = Node.new(nodedata)
    self._nodes[nodeid] = node
    self._newnodes[nodeid] = true
    return node
end

-- getters
function Device:getMAC()
    return self._table.MAC
end

function Device:getConfigTimestamp()
    return self._table.config_timestamp
end

function Device:getKeepalive()
    return self._table.keepalive
end

function Device:getLastConnection()
    return self._table.last_connection
end

-- TODO: this function is a bit silly and only implemented to match
-- current protocol's implementation. Remove when protocol is re-worked and
-- lite restrictions' cache is merged with dao
function Device:getSensorsByType()
    local sensors = {}
    local actuators = {}
    for _, sensor in pairs(self._sensors) do
        if sensor:isSensor() then
            table.insert(sensors, sensor)
        elseif sensor:isActuator() then
            table.insert(actuators, sensor)
        end
    end
    return sensors, actuators
end

function Device:getNodes()
    local nodes = {}
    for _, node in pairs(self._nodes) do
        if node:isServerDefined() then
            table.insert(nodes, node)
        end
    end
    return nodes
end

function Device:getActiveSensors()
    local active = {}
    for _, sensor in pairs(self._sensors) do
        if sensor:isActive() then
            table.insert(active, sensor._sensorid)
        end
    end
    return active
end

function Device:getActiveNodes()
    local active = {}
    for _, node in pairs(self._nodes) do
        if node:isActive() then
            table.insert(active, node._nodeid)
        end
    end
    return active
end

-- setters
function Device:setLastConnection(v)
    self:set('last_connection', 'number', v)
end

function Device:setConfigTimestamp(v)
    self:set('config_timestamp', 'number', v)
end

function Device:toUpdateQuery()
    local statements = {}

    for sensorid, _ in pairs(self._sensors) do
        if self._newsensors[sensorid] then
            table.insert(statements, self._sensors[sensorid]:toInsertQuery())
        else
            local update = self._sensors[sensorid]:toUpdateQuery()
            if update then
                table.insert(statements, update)
            end
        end
    end

    for nodeid, _ in pairs(self._nodes) do
        if self._newnodes[nodeid] then
            table.insert(statements, self._nodes[nodeid]:toInsertQuery())
        else
            local update = self._nodes[nodeid]:toUpdateQuery()
            if update then
                table.insert(statements, update)
            end
        end
    end

    if self:isDirty() then
        table.insert(
            statements,
            query.UPDATE('device'):
                SET(self:dirtyFields()):
                WHERE('vendor_ID', '=', self._vendorid):
                AND('device_ID', '=', self._deviceid)
        )
    end
    return statements
end

function Device:commit()
    local updates = self:toUpdateQuery()
    if not next(updates) then
        return true
    end

    NOTIFYF("Committing changes of device %s.%s", self._vendorid, self._deviceid)

    if query.AS_TRANSACTION(updates) then
        for _, sensor in pairs(self._sensors) do
            sensor:clearDirtyFields()
        end
        self._newsensors = {}
        for _, node in pairs(self._nodes) do
            node:clearDirtyFields()
        end
        self._newnodes = {}
        self:clearDirtyFields()
        return true
    end

    return false
end


return Device
