-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

--- A class to build insert queries
-- This class makes use of the builder pattern to allow a fluent and flexible
-- query building process.
-- @classmod Insert

local convert = require "db/convert"
local checkQuery = require "db/check_query"

local Insert = {}
Insert.__index = Insert

--- Constructor to create a new object of this class.
-- @treturn Insert new object.
function Insert.create()
    local insertObj = {}
    setmetatable(insertObj, Insert)

    insertObj._begin = ''
    insertObj._commit = ''
    insertObj._values = {}
    insertObj._insert = 'INSERT'
    return insertObj
end

function Insert.createReplace()
    local self = Insert.create()
    self._insert = 'REPLACE'
    return self
end

--- Set the table where the insertion should happen.
-- This defines the part of the query after the INTO keyword. This function has
-- to be called at least once to make the query valid.
-- @tparam string tableName name of the table where rows should be deleted.
-- @treturn Insert reference to the calling object.
function Insert:INTO(tableName)
    assert(type(tableName) == "string")
    self._table = tableName
    return self
end

--- Set the column names that should be set when inserting a new row.
-- In a query this is the optional part (column1,column2,..) before the keyword
-- VALUES. Make sure it has the same amount of parameters as passed to VALUES.
-- @tparam string ... columns which should be set with values.
-- @treturn Insert reference to the calling object.
function Insert:COLUMNS(...)
    self._columns = {...}
    return self
end

--- Sets the values that populate the row which will be inserted.
-- If COLUMNS was not called, the parameters have to be the same amount as the
-- target table's columns. This function can be called multiple times to add
-- multiple rows in one statement.
-- @tparam string ... values that should be set for the row.
-- @treturn Insert reference to the calling object.
function Insert:VALUES(...)
    table.insert(self._values, {...})
    return self
end

--- Allows mapping column names to values using a Lua table.
-- This adds the optional COLUMNS parameter to the query.
-- @tparam table valueTable a mapping of columns to values
-- @treturn Insert reference to the calling object.
function Insert:VALUE_TABLE(valueTable)
    assert(type(valueTable) == 'table')
    local columns = {}
    local values = {}
    for column, value in pairs(valueTable) do
        table.insert(columns, column)
        table.insert(values, value)
    end

     table.insert(self._values, values)
     self._columns = columns
     return self
end

--- Checks if the query and its parameters are valid.
-- @treturn boolean true if the query is ok, else false.
function Insert:check()
    local checked = true
    for _, values in ipairs(self._values) do
        checked = checked and checkQuery.values(values)
    end
    return checked and (not self._columns or checkQuery.columns(self._columns))
end

--- Creates a string representation of the Insert object.
-- @treturn string string representation of the query if it is valid, else nil.
function Insert:toString()
    assert(self._table)
    assert(#self._values > 0)
    if not self:check() then
        return nil
    end

    local resultTable = {self._insert, "INTO", self._table}
    if self._columns then
        for _,valueList in ipairs(self._values) do
            assert(#self._columns == #valueList)
        end
        table.insert(resultTable, string.format("(%s)",
            table.concat(self._columns, ",")))
    end

    table.insert(resultTable, "VALUES")
    local recordList = {}
    for _, valueList in ipairs(self._values) do
        local record = {}
        for _, value in ipairs(valueList) do
            table.insert(record, convert.toDBVal(value))
        end
        table.insert(recordList, string.format("(%s)", table.concat(record, ",")))
    end
    table.insert(resultTable, table.concat(recordList, ","))

    return string.format("%s%s;%s", self._begin,
        table.concat(resultTable, " "), self._commit)
end

--- Executes the query.
-- @treturn boolean true if the query was successfully executed, else false.
function Insert:execute()
    local queryStr = self:toString()

    if not queryStr then
        return false
    end

    local insertions = dbExec(queryStr)
    return insertions > 0, insertions
end

return Insert
