-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

--- An interface to build queries for database access and execute them.
-- This module allows building queries on the fly using the builder pattern.
-- Objects perform various checks to validate a query and offer an easy to read
-- interface.
-- @module query

local Select = require "db/select"
local Insert = require "db/insert"
local Update = require "db/update"
local Delete = require "db/delete"

local query = {}

--- Create a Select object to build a SELECT query.
-- @tparam[opt] string ... columns that should be selected.
-- @treturn Select object for further query building.
function query.SELECT(...)
    local selectObj = Select.create()

    if ... then
        selectObj:COLUMNS(...)
    else
        selectObj:COLUMNS("*")
    end

    return selectObj
end

--- Create an Insert object to build an INSERT INTO query.
-- @tparam string tableName name of the target table.
-- @treturn Insert object for further query building.
function query.INSERT_INTO(tableName)
    local insertObj = Insert.create()
    insertObj:INTO(tableName)
    return insertObj
end

function query.REPLACE_INTO(tableName)
    local insertObj = Insert.createReplace()
    insertObj:INTO(tableName)
    return insertObj
end

--- Create Update object to build an UPDATE query.
-- @tparam string tableName name of the target table.
-- @treturn Update object for further query building.
function query.UPDATE(tableName)
    local updateObj = Update.create()
    updateObj:TABLE(tableName)
    return updateObj
end

--- Create Delete object to build a DELETE query.
-- @treturn Delete object for further query building.
function query.DELETE()
    return Delete.create()
end

--- Executes the given query objects as a single transaction.
-- @tparam table queryObjs list of query objects to be executed.
-- @treturn boolean true if the execution of the queries succeeded, else false.
function query.AS_TRANSACTION(queryObjs)
    local statements = {}

    for _, obj in ipairs(queryObjs) do
        if not obj:check() then
            return false
        end
        table.insert(statements, obj:toString())
    end

    local ret = dbExec(table.concat(statements, ' '), 1) -- as transaction

    if ret < 0 then
        dbExec('ROLLBACK;')
    end

    return ret >= 0, ret
end

--- Creates an iterator for queries that can be used in for loops.
-- Each iteration returns a new row.
-- @tparam Select queryObj the object to be iterated
-- @treturn function iteration function for queries
function query.iterator (queryObj)
    return function ()
        return queryObj:next()
    end
end

return query
