-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local checkQuery = require "db/check_query"
local Where = require "db/where"

local Select = {}
Select.__index = Select

--- Constructor to create a new object of this class.
-- @treturn Select new object.
function Select.create()
    local selectObj = {}
    setmetatable(selectObj,Select)
    return selectObj
end

--- Specify the columns that should be read.
-- This defines the part of the query after the SELECT keyword indicating
-- the columns that will be selected.
-- Using this function is optional
-- @tparam string ... Names of selected columns
-- @treturn Select reference to the calling object.
function Select:COLUMNS(...)
    self._columns = {...}
    return self
end

--- Set the table from which should be read.
-- @tparam string tableName
-- @treturn Select reference to the calling object.
function Select:FROM(tableName)
    assert(type(tableName) == "string")
    self._from = tableName
    return self
end

--- Start a conditional part of the query.
-- This is optional and not required for a valid query.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Select reference to the calling object.
function Select:WHERE(columnName, operator, value)
    assert(not self._where)
    self._where = Where.create(columnName, operator, value)
    return self
end

--- Add a condition for the query.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Select reference to the calling object.
function Select:CONDITION(columnName, operator, value)
    assert(self._where)
    self._where:CONDITION(columnName, operator, value)
    return self
end

--- Concatenate two conditions with AND.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Select reference to the calling object.
function Select:AND(columnName, operator, value)
    assert(self._where)
    self._where:AND(columnName, operator, value)
    return self
end

--- Concatenate two conditions with OR.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Select reference to the calling object.
function Select:OR(columnName, operator, value)
    assert(self._where)
    self._where:OR(columnName, operator, value)
    return self
end

--- Checks if the query and its parameters are valid.
-- @treturn boolean true if the query is ok, else false.
function Select:check()
    return (not self._where or self._where:check()) and
       checkQuery.columns(self._columns)
end

--- Creates a string representation of the Select object.
-- @treturn string string representation of the query if it is valid, else nil.
function Select:toString()
    assert(self._columns and #self._columns > 0)
    assert(self._from)
    if not self:check() then
        return nil
    end

    local resultTable = {
        "SELECT",
        table.concat(self._columns, ","),
        "FROM",
        self._from
    }
    if self._where then
        table.insert(resultTable, self._where:toString())
    end

    return string.format("%s;", table.concat(resultTable, " "))
end

--- Executes the query.
-- @treturn boolean true if the query was successfully executed, else false.
function Select:execute()
    local result = {}
    local queryStr = self:toString()

    if not queryStr then
        return false
    end
    result.data = dbQueryRow(queryStr)
    result.count = #result.data
    return result
end

--- Prepare the query stored in this instance.
function Select:prepare()
    local queryStr = self:toString()
    if not queryStr then
        return false
    end

    self._prepared = dbQueryPrepare (queryStr)
    if self._prepared < 0 then
        self._prepared = nil
        return false
    end
    return true
end

--- Get next results row.
function Select:next()
    assert (self._prepared)
    return dbQueryNext (self._prepared)
end

--- Finish prepared query
function Select:finish()
    assert (self._prepared)
    dbQueryFree (self._prepared)
end

return Select
