-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local Where = require "db/where"
local checkQuery = require "db/check_query"
local convert = require "db/convert"

local Update = {}
Update.__index = Update

--- Constructor to create a new object of this class.
-- @treturn Update new object.
function Update.create()
    local updateObj = {}
    setmetatable(updateObj, Update)

    return updateObj
end

--- Set the table from which should be updated.
-- @tparam string tableName
-- @treturn Update reference to the calling object.
function Update:TABLE(tableName)
    assert(tableName)
    self._table = tableName
    return self
end

-- Set the values that should be altered. The Parameter has
-- to be a table with key value pairs. The key specifies the
-- column to be updated and the values the new value of the
-- column. In the query this defines the column1=value,...
-- part after the SET keyword.


--- Set the values that should be altered.
-- In the query this defines the column1=value,... part after
-- the SET keyword.
-- @tparam table nameValueTable a table with key value pairs.
-- The key specifies the column to be updated and the values
-- the new value of the column.
-- @treturn Update reference to the calling object.
function Update:SET(nameValueTable)
    assert(type(nameValueTable) == "table")
    self._set = nameValueTable
    return self
end

function Update:SETV(nameValueTable)
    assert(type(nameValueTable) == "table")
    self._set = nameValueTable
    self._var = true
    return self
end

--- Start a conditional part of the query.
-- This is optional and not required for a valid query.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Update reference to the calling object.
function Update:WHERE(columnName, operator, value)
    assert(not self._where)
    self._where = Where.create(columnName, operator, value)
    return self
end

--- Add a condition for the query.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Update reference to the calling object.
function Update:CONDITION(columnName, operator, value)
    assert(self._where)
    self._where:CONDITION(columnName, operator, value)
    return self
end

--- Concatenate two conditions with AND.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Update reference to the calling object.
function Update:AND(columnName, operator, value)
    assert(self._where)
    self._where:AND(columnName, operator, value)
    return self
end

--- Concatenate two conditions with OR.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Update reference to the calling object.
function Update:OR(columnName, operator, value)
    assert(self._where)
    self._where:OR(columnName, operator, value)
    return self
end

--- Checks if the query and its parameters are valid.
-- @treturn boolean true if the query is ok, else false.
function Update:check()
    return (not self._where or self._where:check()) and
        checkQuery.columnValuePairs(self._set)
end

--- Creates a string representation of the Update object.
-- @treturn string string representation of the query if it is valid, else nil.
function Update:toString()
    assert(self._table)
    assert(self._set)
    if not self:check() then
        return nil
    end

    local resultTable = { "UPDATE", self._table, "SET" }
    local setTable = {}
    if not self._var then
        for k, v in pairs(self._set) do
            table.insert(setTable, string.format("%s=%s", k, convert.toDBVal(v)))
        end
    else
        for k, v in pairs(self._set) do
            table.insert(setTable, string.format("%s=%s", k, tostring(v)))
        end
    end

    table.insert(resultTable, table.concat(setTable, ","))
    if self._where then
        table.insert(resultTable, self._where:toString())
    end

    return string.format("%s;", table.concat(resultTable, " "))
end

--- Executes the query.
-- @treturn boolean true if the query was successfully executed, else false.
function Update:execute()
    local queryStr = self:toString()

    if not queryStr then
        return false
    end

    local updates = dbExec(self:toString())
    return updates >= 0, updates
end

return Update
