-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local convert = require "db/convert"
local checkQuery = require "db/check_query"

local Where = {}
Where.__index = Where

local CONDITION_OP = {
    ["="] = true,
    [">"] = true,
    ["<"] = true,
    [">="] = true,
    ["<="] = true,
    ["<>"] = true,
    ["!="] = true,
    ["LIKE"] = true
}

--- Start a conditional part of the query.
-- This is optional and not required for a valid query.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Where reference to the calling object.
function Where.create(columnName, operator, value)
    local whereObj = {}
    setmetatable(whereObj,Where)

    whereObj._statementList = {"WHERE"}

    if columnName then
        whereObj:CONDITION(columnName, operator, value)
    end

    return whereObj
end

--- Add a condition for the query.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Where reference to the calling object.
function Where:CONDITION(columnName, operator, value)
    assert(value)
    assert(type(columnName) == "string")
    assert(type(operator) == "string")
    assert(CONDITION_OP[operator])

    table.insert(self._statementList, {columnName, operator, value})

    return self
end

--- Concatenate two conditions with AND.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Where reference to the calling object.
function Where:AND(columnName, operator, value)
    table.insert(self._statementList, "AND")

    if columnName then
        self:CONDITION(columnName, operator, value)
    end

    return self
end

--- Concatenate two conditions with OR.
-- WHERE has to be called before this function.
-- @tparam string columnName name of a column as first operand.
-- @tparam string operator operator for the condition.
-- @tparam string|number value value as second operator.
-- @treturn Where reference to the calling object.
function Where:OR(columnName, operator, value)
    table.insert(self._statementList, "OR")

    if columnName then
        self:CONDITION(columnName, operator, value)
    end

    return self
end

--- Checks if the query and its parameters are valid.
-- @treturn boolean true if the query is ok, else false.
function Where:check()
    for _, value in ipairs(self._statementList) do
        if type(value) == "table" and
            (not checkQuery.singleColumn(value[1]) or
            not checkQuery.singleValue(value[3])) then
            return false
        end
    end
    return true
end

--- Creates a string representation of the Where object.
-- @treturn string string representation of the query if it is valid, else nil.
function Where:toString()
    local resultList = {}
    for _, value in ipairs(self._statementList) do
        if type(value) == "table" then
            table.insert(resultList, string.format("%s %s %s", value[1], value[2], convert.toDBVal(value[3])))
        else
            table.insert(resultList, value)
        end
    end
    return table.concat(resultList, " ")
end

return Where
