-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

--[[ METADATA
   DeviceName: CoAp sensor
   Description: Receives data from remote CoAp sensor
   Version: 1.0.0
   Author: Emutex LTD
   AuthorURI: http://emutex.com
   Type: Interface
   Platform: Generic
--]]

local utils = require 'utils/utils'
local tags = require 'utils/tags'
local deviceControl = require 'utils/device_control'
local config = require 'utils/config'
local query = require 'db/query'

libOpen("coap.so.1.0")

local params = { ... }

local vendor = params[1]
local deviceId = params[2]
local sensorId = params[3]

function string:split( inSplitPattern, outResults )
    if not outResults then
        outResults = { }
    end
    local theStart = 1
    local theSplitStart, theSplitEnd = string.find( self, inSplitPattern, theStart )
    while theSplitStart do
        table.insert( outResults, string.sub( self, theStart, theSplitStart-1 ) )
        theStart = theSplitEnd + 1
        theSplitStart, theSplitEnd = string.find( self, inSplitPattern, theStart )
    end
    table.insert( outResults, string.sub( self, theStart ) )
    return outResults
end

local function makeSensorID (device, sensor)
    return device .. "_" .. sensor
end

local function processAnounce(device, value)
    local params = paramDecode(value.payload, config.INTERNAL_PARSER)

    if not params then
        ERRORF ("Unparsable CoAp announce")
        return
    end
    if not params.uid then
        ERRORF ("No uid found for CoAp announce")
        return
    end

    local sensorID = makeSensorID (device, params.uid)

    local hwtype = 'localsensor'
    local hwaddr = ''
    params.type = 2
    if value.type == "actuator" then
        hwtype = 'coapActuator'
        params.type = 1
        hwaddr = table.concat({
            '1=' .. value.addr,
            '2=' .. device,
            '3=' .. params.uid
        }, ',')
    end

    local values = {
        vendor_ID = vendor,
        device_ID = deviceId,
        sensor_ID = sensorID,
        name = params.name or params.uid,
        type = params.type,
        hwaddr = hwaddr,
        hwtype = hwtype,
        scale = 1,
        offset = 0,
        max = tonumber(params.max) or 100,
        min = tonumber(params.min) or 0,
        origin = 1,
        timestamp = os.time(),
        units = params.unit or '',
        mobile = 0,
        latitude = 0,
        longitude = 0,
        active = 1,
        sampleRate = tonumber(params.rate) or 10000,
    }

    -- check if the sensor already exists
    local result = query.
        SELECT():
        FROM('sensor'):
        WHERE('sensor_ID', '=', sensorID):
        execute()

    -- insert or update sensor details
    if result.count ~= 0 then
        NOTIFYF("Sensor %q already exists, updating...", sensorID)
        result = query.
            UPDATE ('sensor'):
            SET (values):
            WHERE ('sensor_ID', '=', sensorID):
            execute ()
    else
        NOTIFYF ('Discovered new CoAp sensor %q', sensorID)
        result = query.
            INSERT_INTO('sensor'):
            VALUE_TABLE(values):
            execute()
    end

    if result then
        NOTIFYF ("   Uid    = %q", params.uid)
        NOTIFYF ("   Name   = %q", values.name)
        NOTIFYF ("   Type   = %q", values.type)
        NOTIFYF ("   HwAddr = %q", values.hwaddr)
        NOTIFYF ("   Min    = %d", values.min)
        NOTIFYF ("   Max    = %d", values.max)
        NOTIFYF ("   Unit   = %q", values.units)
        NOTIFYF ("   Rate   = %d", values.sampleRate)
    else
        ERRORF('Failed to insert/update CoAp sensor %q', sensorID)
    end

end

local function processRequest(tag, value)
    NOTIFYF ("Proccesing CoAp request: %q", value)

    local params = paramDecode(value, config.INTERNAL_PARSER)

    if not params.code or not params.addr or not params.uri or not params.payload then
        ERRORF ("Bad CoAp request")
        NOTIFYF ("   Code    = %s", params.code    or "Undefined!")
        NOTIFYF ("   Address = %s", params.addr    or "Undefined!")
        NOTIFYF ("   URI     = %s", params.uri     or "Undefined!")
        NOTIFYF ("   Payload = %s", params.payload or "Undefined!")
        if not params.code or not params.addr or not params.uri then
            return
        end
    end

    local uriParts = params.uri:split("/")

    local action = uriParts[2]
    local device = uriParts[3]
    params.type  = uriParts[4]

    if not action or not device then
        WARNF ("CoAp URI not valid")
        return
    end

    if action == "announce" then
        NOTIFYF ("Announce for: %q", device)
        processAnounce(device, params)
    elseif action == "update" then
        local sensor = uriParts[5]
        if not sensor then
            ERRORF ("Update CoAp URI not valid")
            return
        end
        local sensorID = makeSensorID (device, sensor)
        NOTIFYF ("New value for %q", sensorID)
        sensorSet (sensorID, params.payload)
    elseif action == "set" then
        WARNF ("Ignored 'set' action for %q", sensorID)
    else
        WARNF ("Ignored unknown action %q", action)
    end

end
eventHandler ("coap.request", processRequest)

coapOpen()

local function update()
    return 0
end

deviceControl.loopSensorThread(vendor, deviceId, sensorId, update, 60000)
