-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

--[[ METADATA
   DeviceName: ESP3 receiver
   Description: EnOcean Serial Protocol 3 receiver (needed for using 'ESP3 input' and 'ESP3 node')
   Version: 1.0.0
   Author: Emutex LTD
   AuthorURI: http://emutex.com
   Type: Interface
   Platform: Generic
   Param1: STR "Serial port"
--]]

-- Removing this parameter from the metadata for now.
-- TODO: Re-add it when discovery is supported by the enocean module
--   Param2: STR "Discovery"   "true"

libOpen('libserial.so.1.0')
libOpen('fieldbus.so.1.0')
libOpen('esp3.so.1.0')

local deviceControl = require 'utils/device_control'
local utils = require 'utils/utils'
local query = require 'db/query'

local params = { ... }
local vendorID = params[1]
local deviceID = params[2]
local sensorID = params[3]
local serial = params[4]
local discovery = utils.tobool(params[5] or 'false')

local DISCOVER_TAG = '_esp3.discover'

local registeredNodes = {}
local registeredInputs = {}

if not serial or discovery == nil then
    ERRORF ('%s invalid hwaddr for %q', utils.getScriptName (), sensorID)
    return
end

local espHandle = esp3Open (serial)
if espHandle < 0 then
    ERRORF ('Failed to open esp3 serial port %q', serial)
    return
end

-- Function to register a node
local function registerNodeCB (tag, param)
    local paramList = paramDecode (param, 'csv')

    -- check if message is correct
    if not paramList.name or not paramList.addr then
        ERRORF ('Esp3 registerNode invalid message %q', param)
        return
    end

    -- node was already registered
    if registeredNodes[paramList.name] then
        return
    end

    -- register the node
    registeredNodes[paramList.name] = esp3RegisterNode (espHandle,
        paramList.addr, paramList.name)
end

--  Function to register a input.  This only works if the
--  device has already been discovered or registered so it is a good to idea to keep
--  calling this function until it succeeds.
local function registerInputCB (tag, param)
    local paramList = paramDecode (param, 'csv')

    -- check if message contains all fields
    if not paramList.node or not paramList.index or not paramList.tag then
        ERRORF ('Esp3 registerInput invalid message %q', param)
        return
    end

    -- check if the input was registered already
    if registeredInputs[paramList.tag] then
        return
    end

    -- check if the node was registered already
    local nodeHnd = registeredNodes[paramList.node]
    if not nodeHnd then
        WARNF ('Esp3 registerInput to unknown node %q',
            tostring (paramList.node))
        return
    end

    local result = esp3DefineInput (nodeHnd, paramList.index,
        paramList.tag)
    if result then
        registeredInputs[paramList.tag] = true
    end
end

--  Function to send command directly to a remote enocean device.
local function outputCB (tag, param)
    local paramList = paramDecode (param, 'csv')

    -- check if message contains all fields
    if not paramList.node or not paramList.command then
        ERRORF ('Esp3 output invalid message %q', param)
        return
    end

    -- check if node was already registered
    local nodeHnd = registeredNodes[paramList.node]
    if not nodeHnd then
        WARNF ('Esp3 output command to unknown device %q',
            tostring(paramList.node))
        return
    end

    esp3SendData (espHandle, nodeHnd, paramList.command,
        string.len(paramList.command))
end

-- callback to discover enocean nodes
local function discoveryFunc (tag, param)
    local paramList = paramDecode (param, 'csv')

    -- check if message contains all fields
    if not paramList.address or not paramList.text then
        ERRORF ('Esp3 discovery invalid message %q', param)
        return
    end

    local nodeHwaddr = paramEncode({
        paramList.text,
        paramList.address
    })
    local nodeHwtype = 'esp3-node'

    -- check if the sensor for this node already exists
    local result = query.
        SELECT('COUNT(sensor_ID) as count'):
        FROM('sensor'):
        WHERE('hwaddr', '=', nodeHwaddr):
        AND('hwtype', '=', nodeHwtype):
        execute()

    if result.count ~= 0 then
        return
    end

    result = query.
        INSERT_INTO('sensor'):
        COLUMN_VALUES({
            vendor_ID = vendorID,
            device_ID = deviceID,
            sensor_ID = paramList.address,
            name = string.format('Esp3 Node %s', paramList.address),
            type = 0,
            hwaddr = nodeHwaddr,
            hwtype = nodeHwtype,
            scale = 1,
            offset = 0,
            max = 100,
            min = 0,
            origin = 1,
            timestamp = os.time(),
            units = 'units',
            mobile = 0,
            latitude = 0,
            longitude = 0,
            active = 1,
            sampleRate = 60000,
        }):
        execute()
    if result then
        NOTIFYF ('Discovered node %q with address %q', paramList.text,
            paramList.address)
    else
        ERRORF ('Failed to insert node %q with address %q', paramList.text,
            paramList.address)
    end
end

local function updateNetwork ()
    if discovery then
        esp3Discover(espHandle,DISCOVER_TAG)
    end

    return utils.tableLength(registeredNodes)
end

eventHandler('_esp3.output',outputCB)
eventHandler('_esp3.register.node',registerNodeCB)
eventHandler('_esp3.register.input',registerInputCB)
if discovery then
    eventHandler(DISCOVER_TAG, discoveryFunc)
end

deviceControl.loopSensorThread(vendorID, deviceID, sensorID, updateNetwork, 60000)

if discovery then
    eventHandlerClear (DISCOVER_TAG)
end
eventHandlerClear ('_esp3.register.input')
eventHandlerClear ('_esp3.register.node')
eventHandlerClear ('_esp3.output')
esp3Close (espHandle)
espHandle = nil
