-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

--[[ METADATA
   DeviceName: XBee
   Description: XBee
   Version: 1.0.0
   Author: Emutex LTD
   AuthorURI: http://emutex.com
   Type: Interface
   Platform: Generic
   Param1:  STR "Serial port"
   Param2:  STR "Discovery"   "true"
--]]

libOpen('libserial.so.1.0')
libOpen('fieldbus.so.1.0')
libOpen('xbee.so.1.0')

local deviceControl = require 'utils/device_control'
local utils = require 'utils/utils'
local query = require 'db/query'

local params = { ... }
local vendorID = params[1]
local deviceID = params[2]
local sensorID = params[3]
local serial = params[4]
local discovery = utils.tobool(params[5] or 'true')

local DISCOVER_TAG = 'xbee.discover'

local registeredNodes = {}
local registeredInputs = {}

if not serial or discovery == nil then
    ERRORF ('%s invalid hwaddr for %q', utils.getScriptName (), sensorID)
    return
end

local xbeeHandle = xbeeOpen(serial)
if xbeeHandle < 0 then
    ERRORF ('Failed to open xbee serial port %q', serial)
    return
end

-- Function to register a node
local function registerNodeCB (tag, param)
    local paramList = paramDecode (param, 'csv')

    -- check if message is correct
    if not paramList.name or not paramList.addr then
        ERRORF ('XBEE registerNode invalid message %q', param)
        return
    end

    -- node was already registered
    if registeredNodes[paramList.name] then
        return
    end

    -- register the node
    registeredNodes[paramList.name] = xbeeRegisterNode (xbeeHandle,
        paramList.addr, paramList.name)
end

--  Function to register a digital or analog XBee input.  This only works if the
--  device has already been discovered or registered so it is a good to idea to keep
--  calling this function until it succeeds.
local function registerInputCB (tag, param)
    local paramList = paramDecode (param, 'csv')
    paramList.digital = tonumber(paramList.digital)

    -- check if message contains all fields
    if not paramList.node or not paramList.index or not paramList.tag
            or not paramList.digital then
        ERRORF ('XBEE registerInput invalid message %q', param)
        return
    end

    -- check if the input was registered already
    if registeredInputs[paramList.tag] then
        return
    end

    -- check if the node was registered already
    local nodeHnd = registeredNodes[paramList.node]
    if not nodeHnd then
        WARNF ('XBEE registerInput to unknown device %q',
            tostring(paramList.node))
        return
    end

    local result = false
    if paramList.digital == 0 then
        result = xbeeDefineAnalogInput (nodeHnd, paramList.index,
                                        paramList.tag)
    else
        result = xbeeDefineDigitalInput (nodeHnd, paramList.index,
                                         paramList.tag)
    end

    if result then
        registeredInputs[paramList.tag] = true
    end
end

--  Function to send text directly to a remote XBee device.  For example a
--  TOSR04 relay board.
local function outputCB (tag, param)
    local paramList = paramDecode (param, 'csv')

    -- check if message contains all fields
    if not paramList.node or not paramList.command then
        ERRORF ('XBEE output invalid message %q', param)
        return
    end

    -- check if node was already registered
    local nodeHnd = registeredNodes[paramList.node]
    if not nodeHnd then
        WARNF ('XBEE output command to unknown device %q',
            tostring(paramList.node))
        return
    end

    xbeeSendData (nodeHnd, paramList.command)
end

-- callback to discover xbees nodes
local function discoveryFunc (tag, param)
    local paramList = paramDecode (param, 'csv')

    -- check if message contains all fields
    if not paramList.address or not paramList.text then
        ERRORF ('XBEE discovery invalid message %q', param)
        return
    end

    local nodeHwaddr = paramEncode({
        paramList.text,
        paramList.address
    })
    local nodeHwtype = 'xbee-node'

    -- check if the sensor for this ndoe already exists
    local result = query.
        SELECT('COUNT(sensor_ID) as count'):
        FROM('sensor'):
        WHERE('hwaddr', '=', nodeHwaddr):
        AND('hwtype', '=', nodeHwtype):
        execute()

    if result.count ~= 0 then
        return
    end

    result = query.
        INSERT_INTO('sensor'):
        COLUMN_VALUES({
        vendor_ID = vendorID,
        device_ID = deviceID,
        sensor_ID = paramList.address,
        name = string.format('Xbee Node %s', paramList.address),
        type = 0,
        hwaddr = nodeHwaddr,
        hwtype = nodeHwtype,
        scale = 1,
        offset = 0,
        max = 100,
        min = 0,
        origin = 1,
        timestamp = os.time(),
        units = 'units',
        mobile = 0,
        latitude = 0,
        longitude = 0,
        active = 1,
        sampleRate = 60000,
        }):
        execute()
    if result then
        NOTIFYF ('Discovered node %q with address %q', paramList.text,
            paramList.address)
    else
        ERROR('Failed to insert node %q with address %q', paramList.text,
            paramList.address)
    end
end

local function updateNetwork ()
    if discovery then
        xbeeDiscover(xbeeHandle,DISCOVER_TAG)
    end

    return utils.tableLength(registeredNodes)
end

eventHandler('_xbeeoutput',outputCB)
eventHandler('_xbeeregisternode',registerNodeCB)
eventHandler('_xbeeregisterinput',registerInputCB)
if discovery then
    eventHandler(DISCOVER_TAG, discoveryFunc)
end

xbeeGetAllowJoin(xbeeHandle, 0)

deviceControl.loopSensorThread(vendorID, deviceID, sensorID, updateNetwork, 60000)

if discovery then
    eventHandlerClear (DISCOVER_TAG)
end
eventHandlerClear ('_xbeeregisterinput')
eventHandlerClear ('_xbeeregisternode')
eventHandlerClear ('_xbeeoutput')
xbeeClose (xbeeHandle)
xbeeHandle = nil
