-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

--- A class to read the BFM 136 Branch Feeder Monitor through Modbus.
-- @classmod Bfm136

libOpen('libserial.so.1.0')
libOpen('fieldbus.so.1.0')
libOpen('modbus.so.1.0')

local Bfm136 = {}
Bfm136.__index = Bfm136

local BAUDRATE = 19200

--- Creates a new Bfm136 object and connects to the modbus.
-- The default address (factory preset) is 1.
-- @tparam string serialPort path to the serial device file
-- @tparam int address start address of submeters
-- @treturn Bfm136 the created object or nil on failure
function Bfm136.create (serialPort, address)
    local bfmObj = {}
    setmetatable (bfmObj, Bfm136)

    bfmObj.address = tonumber(address)
    if not bfmObj.address then
        ERRORF ('BFM136 invalid address %q', address)
        return nil
    end

    bfmObj.bus = modbusOpen (serialPort, BAUDRATE)
    if bfmObj.bus < 0 then
        ERROR ('BFM136 Failed to open modbus')
        return nil
    end

    bfmObj.submeters = {}

    return bfmObj
end

function Bfm136:openSubmeter (submeter)
    local submeterNum = tonumber(submeter)

    if not submeterNum or submeterNum < 1 or submeterNum > 12 then
        ERRORF ('BFM136 invalid submeter %q', tostring(submeter))
        return false
    end

    local submeterAddr = self.address + submeterNum - 1
    NOTIFYF ('BFM 136 registering submeter %d on address %d', submeterNum, submeterAddr)
    local ret = modbusRegisterDevice (self.bus, submeterAddr)
    if ret < 0 then
        ERRORF ('BFM136 failed to register submeter %q on modbus', submeterNum)
        return false
    end

    self.submeters[submeterNum] = {
        handle = ret,
        registers = {}
    }
    return true
end

function Bfm136:defineRange (submeter, startAddr, endAddr)
    local submeterNum = tonumber (submeter)
    if not self.submeters[submeterNum] and not self:openSubmeter (submeterNum) then
        return false
    end
    local device = self.submeters[submeterNum].handle
    NOTIFYF ('BFM 136 submeter %d registering range %d-%d', submeterNum, startAddr, endAddr)
    modbusDefineInputRegisterRange (device, startAddr, endAddr, 'FIELDBUS_DATA_FMT_INT16_AB', 1, 30000)
end

--- Defines a register on the given submeter.
-- Register addresses can be found in the Datasheet. The most common are:
--   - V1-3     : 256-258
--   - I1-3     : 259-261
--   - kW L1-3  : 262-264
--   - kW Total : 275
-- @tparam int submeter number of the submeter
-- @tparam int registerAddr address of the register to read
function Bfm136:defineRegister (submeter, registerAddr, tag)
    local submeterNum = tonumber (submeter)
    if not self.submeters[submeterNum] and not self:openSubmeter (submeterNum) then
        return false
    end
    local device = self.submeters[submeterNum].handle
    NOTIFYF ('BFM 136 submeter %d defining register %d on tag %q', submeterNum, registerAddr, tag)
    modbusDefineInputRegister (device, registerAddr, tag)
end

function Bfm136:destroy ()
    for _, submeter in pairs (self.submeters) do
        modbusUnregisterDevice (submeter.handle)
    end
    self.submeters = nil
    modbusClose (self.bus)
    self.bus = nil
end

return Bfm136
