-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

--! \brief Provides functions for muxing the IO pins on edison (Arduino Breakout Board)

local muxing = {}

local gpioNameTable = {
    ["0"] = "130",
    ["1"] = "131",
    ["2"] = "128",
    ["3"] = "12",
    ["4"] = "129",
    ["5"] = "13",
    ["6"] = "182",
    ["7"] = "48",
    ["8"] = "49",
    ["9"] = "183",
    ["10"] = "41",
    ["11"] = "43",
    ["12"] = "42",
    ["13"] = "40",
    A0 = "44",
    A1 = "45",
    A2 = "46",
    A3 = "47",
    A4 = "14",
    A5 = "165"
}

function muxing.getGpioNumber(name)
    return gpioNameTable[name]
end

function muxing.writeToFile(fileName, content)
    local f = io.open(fileName, "w")
    if f == nil then
        return false
    end

    f:write(content)
    f:close()
    return true
end

local function exportGPIO(gpioTable)
    for gpio,_ in pairs(gpioTable) do
        if not muxing.writeToFile("/sys/class/gpio/export", gpio) then
            ERROR(string.format("Failed to export gpio '%s'", gpio))
            return false
        end
    end
    return true
end

local function setDirections(gpioTable)
    for gpio,dir in pairs(gpioTable) do
        local fileName = string.format("/sys/class/gpio/gpio%s/direction", gpio)
        if not muxing.writeToFile(fileName, dir) then
            ERROR(string.format("Failed to set direction for gpio '%s' to '%s'", gpio, dir))
            return false
        end
    end

    return true
end

local function setMode(gpioTable)
    for gpio,mode in pairs(gpioTable) do
        local fileName = string.format("/sys/kernel/debug/gpio_debug/gpio%s/current_pinmux", gpio)
        if not muxing.writeToFile(fileName, mode) then
            ERROR("Failed to set mode for gpio '%s' to '%s'", gpio, mode)
            return false
        end
    end

    return true
end

local function startMux()
    local magic214Table = {
        ["214"] = "low"
    }

    local result =
        exportGPIO(magic214Table) and
        setDirections(magic214Table)

    return result
end

local function endMux()
    local magic214Table = {
        ["214"] = "high"
    }
    return setDirections(magic214Table)
end

function muxing.muxPWM(id)
    local pwmMuxTable = {
        ["0"] = { ["251"] = "high", ["219"] = "in" },
        ["1"] = { ["253"] = "high", ["221"] = "in" },
        ["2"] = { ["254"] = "high", ["222"] = "in" },
        ["3"] = { ["257"] = "high", ["225"] = "in" },
        ["4"] = { ["258"] = "high", ["226"] = "in" },
        ["5"] = { ["259"] = "high", ["228"] = "in" }
    }

    local pwmModeTable = {
        ["0"] = { ["12"] = "mode1" },
        ["1"] = { ["13"] = "mode1" },
        ["2"] = { ["182"] = "mode1" },
        ["3"] = { ["183"] = "mode1" },
        ["4"] = { ["41"] = "mode1" },
        ["5"] = { ["43"] = "mode1" }
    }

    if pwmMuxTable[id] == nil then
        ERROR(string.format("Invalid pwm ID '%s'", id))
        return false
    end

    local result =
        exportGPIO(pwmMuxTable[id]) and
        startMux() and
        setDirections(pwmMuxTable[id]) and
        setMode(pwmModeTable[id]) and
        endMux() and
        muxing.writeToFile("/sys/class/pwm/pwmchip0/export", id)
    return result
end

function muxing.muxI2C()
    local i2cMuxTable = {
        ["204"] = "low",
        ["205"] = "low",
        ["14"] = "in",
        ["165"] = "in",
        ["236"] = "low",
        ["237"] = "low",
        ["212"] = "in",
        ["213"] = "in"
    }

    local i2cModeTable = {
        ["28"] = "mode1",
        ["27"] = "mode1"
    }

    local result =
        exportGPIO(i2cModeTable) and
        exportGPIO(i2cMuxTable) and
        startMux() and
        setDirections(i2cMuxTable) and
        setMode(i2cModeTable) and
        endMux()

    return result
end

function muxing.muxADC(id)
    local muxAdcTable = {
        ["0"] = { ["200"] = "high", ["232"] = "low", ["208"] = "in" },
        ["1"] = { ["201"] = "high", ["233"] = "low", ["209"] = "in" },
        ["2"] = { ["202"] = "high", ["234"] = "low", ["210"] = "in" },
        ["3"] = { ["203"] = "high", ["235"] = "low", ["211"] = "in" },
        ["4"] = { ["204"] = "high", ["236"] = "low", ["212"] = "in" },
        ["5"] = { ["205"] = "high", ["237"] = "low", ["213"] = "in" }
    }

    if muxAdcTable[id] == nil then
        ERROR(string.format("Invalid adc id '%s'", id))
        return false
    end

    local result =
        muxing.muxGPIO("13", "in") and
        exportGPIO(muxAdcTable[id]) and
        startMux() and
        setDirections(muxAdcTable[id]) and
        endMux()
    return result
end

function muxing.muxGPIO(pin, direction)
    local outputEnable
    if direction == "in" then
        outputEnable = "low"
    elseif direction == "out" then
        outputEnable = "high"
    else
        ERROR(string.format("Got invalid gpio direction '%s'", direction))
        return false
    end

    local gpioBeforeMode = {
        ["0"] = { ["248"] = outputEnable, ["216"] = "in" },
        ["1"] = { ["249"] = outputEnable, ["217"] = "in" },
        ["2"] = { ["250"] = outputEnable, ["218"] = "in" },
        ["3"] = { ["251"] = outputEnable, ["219"] = "in" },
        ["4"] = { ["252"] = outputEnable, ["220"] = "in" },
        ["5"] = { ["253"] = outputEnable, ["221"] = "in" },
        ["6"] = { ["254"] = outputEnable, ["222"] = "in" },
        ["7"] = { ["255"] = outputEnable, ["223"] = "in" },
        ["8"] = { ["256"] = outputEnable, ["224"] = "in" },
        ["9"] = { ["257"] = outputEnable, ["225"] = "in" },
        ["10"] = { ["263"] = "high", ["240"] = "low" },
        ["11"] = { ["262"] = "high", ["241"] = "low" },
        ["12"] = { ["242"] = "low" },
        ["13"] = { ["243"] = "low" },
        A0 = { ["200"] = "low" },
        A1 = { ["201"] = "low" },
        A2 = { ["202"] = "low" },
        A3 = { ["203"] = "low" },
        A4 = { ["204"] = "low" },
        A5 = { ["205"] = "low" }
    }

    local gpioMode = {
        ["0"] = { ["130"] = "mode0" },
        ["1"] = { ["131"] = "mode0" },
        ["2"] = { ["128"] = "mode0" },
        ["3"] = { ["12"] = "mode0" },
        ["4"] = { ["129"] = "mode0" },
        ["5"] = { ["13"] = "mode0" },
        ["6"] = { ["182"] = "mode0" },
        ["7"] = { ["48"] = "mode0" },
        ["8"] = { ["49"] = "mode0" },
        ["9"] = { ["183"] = "mode0" },
        ["10"] = { ["41"] = "mode0" },
        ["11"] = { ["43"] = "mode0" },
        ["12"] = { ["42"] = "mode0" },
        ["13"] = { ["40"] = "mode0" },
        A0 = { ["44"] = "mode0" },
        A1 = { ["45"] = "mode0" },
        A2 = { ["46"] = "mode0" },
        A3 = { ["47"] = "mode0" },
        A4 = { ["14"] = "mode0" },
        A5 = { ["165"] = "mode0" }
    }

    local gpioAfterMode = {
        ["0"] = { ["130"] = direction },
        ["1"] = { ["131"] = direction },
        ["2"] = { ["128"] = direction },
        ["3"] = { ["12"] = direction },
        ["4"] = { ["129"] = direction },
        ["5"] = { ["13"] = direction },
        ["6"] = { ["182"] = direction },
        ["7"] = { ["48"] = direction },
        ["8"] = { ["49"] = direction },
        ["9"] = { ["183"] = direction },
        ["10"] = { ["258"] = outputEnable, ["226"] = "in", ["41"] = direction },
        ["11"] = { ["259"] = outputEnable, ["227"] = "in", ["43"] = direction },
        ["12"] = { ["260"] = outputEnable, ["228"] = "in", ["42"] = direction },
        ["13"] = { ["261"] = outputEnable, ["229"] = "in", ["40"] = direction },
        A0 = { ["232"] = outputEnable, ["208"] = "in", ["44"] = direction },
        A1 = { ["233"] = outputEnable, ["209"] = "in", ["45"] = direction },
        A2 = { ["234"] = outputEnable, ["210"] = "in", ["46"] = direction },
        A3 = { ["235"] = outputEnable, ["211"] = "in", ["47"] = direction },
        A4 = { ["236"] = outputEnable, ["212"] = "in", ["14"] = direction },
        A5 = { ["237"] = outputEnable, ["213"] = "in", ["165"] = direction }
    }

    if gpioBeforeMode[pin] == nil then
        ERROR(string.format("Got invalid gpio pin '%s'", pin))
        return false
    end

    local result =
        exportGPIO(gpioBeforeMode[pin]) and
        exportGPIO(gpioMode[pin]) and
        exportGPIO(gpioAfterMode[pin]) and
        startMux() and
        setDirections(gpioBeforeMode[pin]) and
        setMode(gpioMode[pin]) and
        setDirections(gpioAfterMode[pin]) and
        endMux()
    return result
end

return muxing
