-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

--- A class to read serial RFID components
-- @classmod Rfid

libOpen ('libserial.so.1.0')

local Queue = require 'utils/queue'
local utils = require 'utils/utils'

local Rfid = {}
Rfid.__index = Rfid

--- Creates a new RFID object.
-- @tparam string port serial port to read from.
-- @tparam integer baudrate baudrate used by the serial device.
-- @tparam[opt] separator string that separates IDs in messages.
-- @treturn Rfid a new instance or nil if the creation failed.
function Rfid.create (port, baudrate, separator)
    separator = separator or '\r\n'
    baudrate = tonumber(baudrate)

    if not port or not baudrate then
        return nil
    end

    local obj = {}
    setmetatable (obj, Rfid)

    obj._port = port
    obj._baudrate = baudrate
    obj._separator = separator
    obj._buf = ''
    obj._idQueue = Queue.create ()

    return obj
end

--- Opens the serial connection to the RFID reader.
-- @treturn boolean returns true on success, else false.
function Rfid:open ()
    self._handle = serialOpen (self._port, self._baudrate)

    return self._handle >= 0
end

--- Reads the next RFID key from the device.
-- Blocks the current lua thread if no data is available.
-- @treturn string RFID key
function Rfid:read ()
    if not self._idQueue:isEmpty () then
        return self._idQueue:popLeft ()
    end

    while true do
        self.buf_ = self._buf .. serialRead (self._handle)

        if self.buf_:match (self._separator) then
            break
        end
    end

    -- remove start and end text characters from buffer
    local pattern = string.format ('[%s]', string.char (0x02,0x03))
    self.buf_ = self.buf_:gsub (pattern, '')
    -- split messages at separator
    local ids = utils.split (self.buf_, self._separator)
    -- check if we received a message only partly
    if ids[1]:len() == ids[#ids]:len () then
        table.insert (ids, '')
    end

    for _, id in ipairs (ids) do
        self._idQueue:pushRight (id)
    end
    self.buf_ = self._idQueue:popRight ()

    local result = self._idQueue:popLeft ()
    return result:sub(1, #result - 2)
end

--- Get the serial port that this device uses.
-- @treturn string serial port
function Rfid:getPort ()
    return self._port
end

--- Get the baudrate that this device uses.
-- @treturn integer serial baudrate
function Rfid:getBaudrate ()
    return self._baudrate
end

--- Closes the serial connection.
function Rfid:close ()
    serialClose (self._handle)
end

return Rfid
