-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local query = require 'db/query'
local tags = require 'utils/tags'
local usensor = require 'utils/sensor'

local self = {}
self.vendors = {}
self.license = require 'lite/license'

self.maxDevices = tonumber(getSystemConfigValue('liteMaxDevs', '2'))
self.maxSensors = tonumber(getSystemConfigValue('liteMaxSensors', '5'))
self.maxActuators = tonumber(getSystemConfigValue('liteMaxActuators', '2'))
self.maxTransactions = tonumber(getSystemConfigValue('liteMaxTrans', '10000'))

local __devices = 'devices'
local __deviceCount = 'device_count'
local __actuators = 'actuators'
local __actuatorCount = 'actuator_count'
local __sensors = 'sensors'
local __sensor_count = 'sensor_count'
local __trans = 'trans'
local __deactivated = 'deactivated'

local function newVendor(vendorID)
    if self.vendors[vendorID] == nil then
        self.vendors[vendorID] = {
            [__devices] = {},
            [__deviceCount] = 0,
        }
    end
end

local function newDevice(vendorID, deviceID)
    newVendor(vendorID)
    if self.vendors[vendorID][__devices][deviceID] == nil then
        self.vendors[vendorID][__devices][deviceID] = {
            [__sensor_count] = 0,
            [__actuatorCount] = 0,
            [__sensors] = {},
            [__actuators] = {},
            [__trans] = 0,
            [__deactivated] = false,
        }
    end
end

local function isAnotherItemAllowed(vendorID, deviceID, itemName, maxItems)
    local item = self.vendors[vendorID][__devices][deviceID][itemName] + 1

    if item <= maxItems then
        self.vendors[vendorID][__devices][deviceID][itemName] = item
        return true
    end

    return false
end

function self.setMQTTServer(mqtt)
    self.mqttServer = mqtt
end

local function selectSensor(vendorID, deviceID)
    return query.
        SELECT():
        FROM('sensor'):
        WHERE('vendor_ID', '=', vendorID):
        AND('device_ID', '=', deviceID):
        AND('removed', '=', 0):
        execute()
end

function self.initSensorsFromDB(vendorID, deviceID)
    local sensors = selectSensor(vendorID, deviceID)

    for _,  row in ipairs(sensors.data) do
        local hwtype = usensor.legacyHwType2HwType(row.hwtype, row.type)
        if usensor.isSensor(hwtype) then
            local count = self.vendors[vendorID][__devices][deviceID][__sensor_count]
            self.vendors[vendorID][__devices][deviceID][__sensor_count] = count + 1
            table.insert(self.vendors[vendorID][__devices][deviceID][__sensors],
                row.sensor_ID)
        end
    end
end

function self.initActuatorsFromDB(vendorID, deviceID)
    local actuators = selectSensor(vendorID, deviceID)

    for _, row in ipairs(actuators.data) do
        local hwtype = usensor.legacyHwType2HwType(row.hwtype, row.type)
        if usensor.isActuator(hwtype) then
            local count = self.vendors[vendorID][__devices][deviceID][__actuatorCount]
            self.vendors[vendorID][__devices][deviceID][__actuatorCount] = count + 1
            table.insert(self.vendors[vendorID][__devices][deviceID][__actuators],
                row.sensor_ID)
        end
    end
end

function self.initFromDB()
    local vendorDevices =
        query.SELECT('vendor_ID', 'device_ID'):
            FROM('device'):
            WHERE('removed', '=', 0):
            execute()

    for _, row in pairs(vendorDevices.data) do
        local vendorID = tostring(row.vendor_ID)
        local deviceID = tostring(row.device_ID)
        newDevice(vendorID, deviceID)
        self.vendors[vendorID][__deviceCount] = self.vendors[vendorID][__deviceCount] + 1

        self.initSensorsFromDB(vendorID, deviceID)
        self.initActuatorsFromDB(vendorID, deviceID)
    end
end

function self.initSensors(vendorID, deviceID, sensors)
    vendorID, deviceID = tostring(vendorID), tostring(deviceID)

    -- if vendorID has a full license no checks are performed
    if not self.license.isLite(vendorID) then
        return 0
    end

    if #sensors > self.maxSensors then
        return #sensors - self.maxSensors
    end

    self.vendors[vendorID][__devices][deviceID][__sensor_count] = #sensors
    return 0
end

function self.initActuators(vendorID, deviceID, actuators)
    vendorID, deviceID = tostring(vendorID), tostring(deviceID)

    -- if vendorID has a full license no checks are performed
    if not self.license.isLite(vendorID) then
        return 0
    end

    if #actuators > self.maxActuators then
        return #actuators - self.maxActuators
    end

    self.vendors[vendorID][__devices][deviceID][__actuatorCount] = #actuators
    return 0
end

function self.isAnotherDeviceAllowed(vendorID, deviceID)
    vendorID, deviceID = tostring(vendorID), tostring(deviceID)

    -- if vendorID has a full license no checks are performed
    if not self.license.isLite(vendorID) then
        return true
    end

    newVendor(vendorID)

    -- if deviceID is already known it will always be allowed
    if self.vendors[vendorID][__devices][deviceID] ~= nil then
        return true
    end

    local numDevices = self.vendors[vendorID][__deviceCount] + 1

    if numDevices <= self.maxDevices then
        self.vendors[vendorID][__deviceCount] = numDevices
        newDevice(vendorID, deviceID)
        return true
    end

    return false
end

function self.removeDevice(vendorID, deviceID)
    vendorID, deviceID = tostring(vendorID), tostring(deviceID)

    if not self.vendors[vendorID][__devices][deviceID] then
        return
    end

    self.vendors[vendorID][__devices][deviceID] = nil

    self.vendors[vendorID][__deviceCount] =
        self.vendors[vendorID][__deviceCount] - 1
end

function self.removeSensor(vendorID, deviceID, sensorID, hwtype, typeIdx)
    vendorID = tostring(vendorID)
    hwtype = usensor.legacyHwType2HwType(hwtype, typeIdx)
    if not usensor.isValidType(hwtype) then
        ERRORF("Invalid hardware type '%s' for %s.%s.%s", hwtype, vendorID, deviceID, sensorID)
        return false
    end

    local type_count_name = __sensor_count
    local type_name = __sensors

    if usensor.isActuator(hwtype) then
        type_count_name = __actuatorCount
        type_name = __actuators
    end

    local sensors = self.vendors[vendorID][__devices][deviceID][type_name]
    for i=#sensors,1,-1 do
        if sensors[i] == sensorID then
            table.remove(sensors, i)
            self.vendors[vendorID][__devices][deviceID][type_count_name] = #sensors
            return
        end
    end
end

function self.isAnotherSensorTypeAllowed(vendorID, deviceID, sensorID, hwtype, typeIdx)
    hwtype = usensor.legacyHwType2HwType(hwtype, typeIdx)
    if not usensor.isValidType(hwtype) then
        ERRORF("Invalid hardware type '%s' for %s.%s.%s", hwtype, vendorID, deviceID, sensorID)
        return false
    end
    vendorID, deviceID = tostring(vendorID), tostring(deviceID)

    -- if vendorID has a full license no checks are performed
    if not self.license.isLite(vendorID) then
        return true
    end

    newDevice(vendorID, deviceID)
    local type_count_name = __sensor_count
    local type_name = __sensors
    local max = self.maxSensors

    if usensor.isActuator(hwtype) then
        type_count_name = __actuatorCount
        type_name = __actuators
        max = self.maxActuators
    end

    local items = self.vendors[vendorID][__devices][deviceID][type_name]
    if items[sensorID] then
        return true
    end

    if isAnotherItemAllowed(vendorID, deviceID, type_count_name, max) then
        items[sensorID] = true
        return true
    end

    return false
end

function self.maxDailyTransactionsReached(vendorID, deviceID)
    vendorID, deviceID = tostring(vendorID), tostring(deviceID)

    -- if vendorID has a full license no checks are performed
    if not self.license.isLite(vendorID) then
        return false
    end

    newDevice(vendorID, deviceID)
    return
        not isAnotherItemAllowed(vendorID, deviceID, __trans, self.maxTransactions)
end

local function setDeviceActive(vendorID, deviceID, active)
    query.UPDATE('device'):
        SET({ active = active }):
        WHERE('vendor_ID', '=', vendorID):
        AND('device_ID', '=', deviceID):
        execute()
end

function self.resetAllDeviceTransactions()
    for vendorID in pairs(self.vendors) do
        for deviceID in pairs(self.vendors[vendorID][__devices]) do
            self.vendors[vendorID][__devices][deviceID][__trans] = 0
            if self.vendors[vendorID][__devices][deviceID][__deactivated] then
                setDeviceActive(vendorID, deviceID, 1)
                self.vendors[vendorID][__devices][deviceID][__deactivated] = false
            end
        end
    end
end

function self.deviceReachedMaxTransactions(vendorID, deviceID)
    vendorID, deviceID = tostring(vendorID), tostring(deviceID)

    if self.vendors[vendorID][__devices][deviceID][__deactivated] then
        -- deviceID already inactive. No need to tell it again.
        return
    end

    local msg = "Device %s.%s has reached its max daily transactions"
    WARN(msg:format(vendorID, deviceID))

    setDeviceActive(vendorID, deviceID, 0)
    self.vendors[vendorID][__devices][deviceID][__deactivated] = true

    local mqttParams = {
        action = 'quota_reached',
    }
    local topic = tags.getClientTag(vendorID, deviceID)
    mqttPublish (self.mqttServer, topic, paramEncodeHeader (mqttParams, 'csv'))
end

return self
