-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local query = require 'db/query'
local tags = require 'utils/tags'
local config = {}

config.CLIENT_BROKER = "mqttClientBroker"
config.CLIENT_PORT = "mqttClientPort"
config.SERVER_BROKER = "mqttServerBroker"
config.SERVER_PORT = "mqttServerPort"
config.TLS_ENABLED = "tlsEnabled"

config.MAILLIST = "mailList"
config.ABS_TIMESTAMPS = "useAbsoluteTimestamps"
config.STORE_AND_FWD = "storeAndForward"
config.DEV_SCRIPTS_METADATA = "devScriptsMetadata"
config.DEVICE_SCRIPTS = "deviceScripts"
config.SCADA_EXPORT = "scadaExport"
config.SCADA_SERVER = "scadaServer"
config.SCADA_PORT = "scadaPort"
config.SCADA_ENABLED = "scadaEnabled"
config.RULES = "gatewayLogic"
config.CLIENT_SEND_BATCH_SIZE = "clientSendBatchSize"
config.CLIENT_RESEND_BATCH_SIZE = "storeAndForwardClientBatchSize"
config.CLIENT_MAX_RESEND_BATCHES = "storeAndForwardClientMaxResendBatches"
config.CLIENT_RESEND_TIMEOUT = "storeAndForwardResendTimeout"
config.PARSER_TYPE = 'parserType'
config.MQTT_PASS = 'mqttPass'
config.HAS_RTC = 'hasRTC'
config.SOFTWARE_UPGRADE_STATUS = 'softwareUpgradeStatus'
config.SOFTWARE_UPGRADE_REQUEST = 'softwareUpgradeRequest'

-- connector stuff
config.CONNECTOR_ENABLED = "connectorEnabled"
config.C8Y_FORWARDING = "c8yForwarding"
config.C8Y_ACCOUNT = "c8yAccount"
config.C8Y_GATEWAY_ID = "c8yGatewayID"
config.JDE_FORWARDING = 'jdeForwarding'
config.JDE_ACCOUNT = 'jdeAccount'
config.JDE_GATEWAY_ID = 'jdeGatewayID'

config.HWADDR_PARSER = 'csv'
config.INTERNAL_PARSER = 'csv'
config.SSL_PORT = 8883

config._loaded = {}

local boolConvertTable = {
    YES = true,
    NO = false,
    Y = true,
    N = false,
    TRUE = true,
    FALSE = false,
    T = true,
    F = false
}

--- Converts input string to a boolean.
-- @tparam string str string value to be converted.
-- @treturn boolean value for input string.
function config.toBool(str)
    return boolConvertTable[str:upper()]
end

--- Construct an ID from vendorID and deviceID.
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string combining inputs IDs into a new ID.
function config.getID(vendorID, deviceID)
    return string.format("%s.%s", vendorID, deviceID)
end

function config.refresh(vendorID, deviceID)
    local id = config.getID(vendorID, deviceID)

    NOTIFYF("Refreshing config of %q", id)

    local resultTable = query.
        SELECT():
        FROM("device_config"):
        WHERE("vendor_ID", "=", vendorID):
        AND("device_ID", "=", deviceID):
        execute()

    if not resultTable then
        return false
    end

    -- don't invalidate the old cache until it's safe. The previous call to the
    -- db will make this thread to sleep.
    config._loaded[id] = {}
    if resultTable.data then
        for _, row in ipairs(resultTable.data) do
            config._loaded[id][row.name] = row.value
        end
    end

    return true
end

function config.clear(vendorID, deviceID)
    local id = config.getID(vendorID, deviceID)
    NOTIFYF("Clearing config of %q", id)
    config._loaded[id] = nil
end

function config.clearAll()
    config._loaded = {}
end

--- Returns config table for the specified vendor and device.
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn Table with the configuration.
function config.get(vendorID, deviceID)
    local id = config.getID(vendorID, deviceID)
    return config._loaded[id]
end

--- Returns specified value for specified vendor and device
--  from the config table.
-- @tparam string vendorID
-- @tparam string deviceID
-- @tparam string name config field name
-- @tparam string defaultValue in case no value on the table
-- @treturn string config value
function config.getValue(vendorID, deviceID, name, defaultValue)
    defaultValue = defaultValue or ""
    local configTable = config.get(vendorID, deviceID)
    assert(configTable, string.format('Failed getting %q entry from config', name))
    return configTable[name] or defaultValue
end

--- Checks if the specified table has the specified config field
-- @tparam Table
-- @tparam string name config field name
-- @treturn boolean true if field exists
local function hasDeviceConfig(resultTable, name)
    if resultTable.count == 0 then
        return false
    end

    for _, row in ipairs(resultTable.data) do
        if row.name == name then
            return true
        end
    end

    return false
end

--- Store the specified config table for the specified
--  vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @tparam Table with the configuration.
-- @treturn boolean true if done.
function config.store(vendorID, deviceID, configTable)
    local id = config.getID(vendorID,deviceID)

    NOTIFYF("Storing new config of %q", id)

    local statements = {}
    for name, value in pairs(configTable) do
        table.insert(statements, query.
            REPLACE_INTO("device_config"):
            COLUMNS("vendor_ID", "device_ID", "name", "value"):
            VALUES(vendorID, deviceID, name, value))
    end

    if not query.AS_TRANSACTION(statements) then
        ERROR("config.store, db transaction failed")
        return false
    end

    -- only update cache if the db has been updated
    -- don't override all table, only update the values from the server
    for name, value in pairs(configTable) do
        config._loaded[id][name] = value
    end

    return true
end

function config.sendValue(vendorID, deviceID, name)
    local csv = {
        action = 'sendconfig',
        name = name
    }

    local msg = paramEncodeHeader(csv, config.getParserType(vendorID, deviceID))
    eventPost(tags.client, msg, true)
end

--- Store specified value into config table for the specified
--  vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @tparam string name config field name
-- @tparam string value Value to be stored
-- @treturn Bool true if done.
function config.storeValue(vendorID, deviceID, name, value)
    local ret = query.
        REPLACE_INTO("device_config"):
        COLUMNS("vendor_ID", "device_ID", "name", "value"):
        VALUES(vendorID, deviceID, name, value):
        execute()

    if ret then
        local id = config.getID(vendorID,deviceID)
        config._loaded[id][name] = value
    end

    return ret
end

--- Gets server broker for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string requested value
function config.getServerBroker(vendorID, deviceID)
    local broker = config.getValue(vendorID, deviceID, config.SERVER_BROKER, "127.0.0.1")
    local port = config.getValue(vendorID, deviceID, config.SERVER_PORT, config.SSL_PORT)
    return broker, tonumber (port)
end

--- Gets client broker for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string requested value
function config.getClientBroker(vendorID, deviceID)
    local broker = config.getValue(vendorID, deviceID, config.CLIENT_BROKER, "broker.ubiworx.com")
    local port = config.getValue(vendorID, deviceID, config.CLIENT_PORT, config.SSL_PORT)
    return broker, tonumber (port)
end

--- Gets mail list for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string requested value
function config.getMailList(vendorID, deviceID)
    return config.getValue(vendorID, deviceID, config.MAILLIST, "")
end

--- Returns if the specified vendor and device uses absolute
--  timestamps
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn boolean requested value
function config.useAbsoluteTimestamps(vendorID, deviceID)
    local result = config.getValue(vendorID, deviceID, config.ABS_TIMESTAMPS, "No")
    return config.toBool(result)
end

--- Returns if the specified vendor and device has store and forward
--- enabled
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn boolean requested value
function config.useStoreAndForward(vendorID, deviceID)
    local result = config.getValue(vendorID, deviceID, config.STORE_AND_FWD, "Yes")
    return config.toBool(result)
end

 --- Set the specified vendor and device device scripts
 -- @tparam string vendorID
 -- @tparam string deviceID
 -- @tparam string value Device scripts
 -- @treturn boolean true if done
function config.setDeviceScripts(vendorID, deviceID, value)
    return config.storeValue(vendorID,deviceID,config.DEVICE_SCRIPTS,value)
end

--- Set the specified vendor and device devices metadata
-- @tparam string vendorID
-- @tparam string deviceID
-- @tparam string value Device metadata
-- @treturn boolean true if done
function config.setDevScriptsMetadata(vendorID, deviceID, value)
    return config.storeValue(vendorID,deviceID,config.DEV_SCRIPTS_METADATA,value)
end

--- Return specified vendor and device devices metadata
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn boolean true if done
function config.getDevScriptsMetadata(vendorID, deviceID)
    return config.getValue(vendorID,deviceID,config.DEV_SCRIPTS_METADATA, nil)
end

--- Returns if the specified vendor and device reports for SCADA
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn boolean requested value
function config.exportForScada(vendorID, deviceID)
    local result = config.getValue(vendorID, deviceID, config.SCADA_EXPORT, "No")
    return config.toBool(result)
end

--- Returns the specified vendor and device assigned SCADA server
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string server URL
-- @treturn number server port
function config.getScadaServer(vendorID, deviceID)
    local server = config.getValue(vendorID, deviceID, config.SCADA_SERVER, "broker.ubiworx.com")
    local port = config.getValue(vendorID, deviceID, config.SCADA_PORT, "8081")
    return server, tonumber (port)
end

--- Returns if the specified vendor and device has SCADA enabled
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn boolean requested value
function config.scadaEnabled(vendorID, deviceID)
    local result = config.getValue(vendorID, deviceID, config.SCADA_ENABLED, "No")
    return config.toBool(result)
end

--- Returns rules for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string requested value
function config.getRules(vendorID, deviceID)
    return config.getValue(vendorID, deviceID, config.RULES, "")
end

--- Returns if the specified vendor and device has connector enabled
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn boolean requested value
function config.connectorEnabled(vendorID, deviceID)
    local result = config.getValue(vendorID, deviceID, config.CONNECTOR_ENABLED, "No")
    return config.toBool(result)
end

--- Returns if the specified vendor and device is forwarding to C8Y
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn boolean requested value
function config.forwardToC8Y(vendorID, deviceID)
    local result = config.getValue(vendorID, deviceID, config.C8Y_FORWARDING, "No")
    return config.toBool(result)
end

--- Returns the C8Y account for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string requested value
function config.getC8YAccount(vendorID, deviceID)
    return config.getValue(vendorID, deviceID, config.C8Y_ACCOUNT, "")
end

--- Returns the C8Y gateway ID for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string requested value
function config.getC8YGatewayID(vendorID, deviceID)
    return config.getValue(vendorID, deviceID, config.C8Y_GATEWAY_ID, "0")
end

--- Returns if the specified vendor and device is forwarding to JDE
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn boolean requested value
function config.forwardToJDE(vendorID, deviceID)
    local result = config.getValue(vendorID, deviceID, config.JDE_FORWARDING, "No")
    return config.toBool(result)
end

--- Returns the JDE account for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string requested value
function config.getJDEAccount(vendorID, deviceID)
    return config.getValue(vendorID, deviceID, config.JDE_ACCOUNT, "")
end

--- Returns the JDE gateway ID for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string requested value
function config.getJDEGatewayID(vendorID, deviceID)
    return config.getValue(vendorID, deviceID, config.JDE_GATEWAY_ID, "0")
end

--- Returns the resended batch size for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn number requested value
function config.getClientResendBatchSize(vendorID, deviceID)
    local ret =
        config.getValue(vendorID, deviceID, config.CLIENT_RESEND_BATCH_SIZE, "100")
    return tonumber(ret)
end

--- Returns the resend batch size for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn number requested value
function config.getClientSendBatchSize(vendorID, deviceID)
    local ret =
        config.getValue(vendorID, deviceID, config.CLIENT_SEND_BATCH_SIZE, "1")
    return tonumber(ret)
end

--- Returns the maximum resend batches for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn number requested value
function config.getClientMaxResendBatches(vendorID, deviceID)
    local ret =
        config.getValue(vendorID, deviceID, config.CLIENT_MAX_RESEND_BATCHES, "2")
    return tonumber(ret)
end

--- Returns resend time out for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn number requested value
function config.getResendTimeout(vendorID, deviceID)
    local ret =
        config.getValue(vendorID, deviceID, config.CLIENT_RESEND_TIMEOUT, '10000')
    return tonumber(ret)
end

--- Returns parser type for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string requested value
function config.getParserType (vendorID, deviceID)
    return config.getValue (vendorID, deviceID, config.PARSER_TYPE, 'csv')
end

--- Returns MQTT password for the specified vendor and device
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn string requested value
function config.getMqttPassword(vendorID, deviceID)
    return config.getValue(vendorID, deviceID, config.MQTT_PASS, '')
end

--- Returns if the specified vendor and device has TLS enabled
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn boolean requested value
function config.isTlsEnabled(vendorID, deviceID)
    local result = config.getValue(vendorID, deviceID, config.TLS_ENABLED, 'Yes')
    return config.toBool(result)
end

--- Returns if the specified vendor and device has RTC
-- @tparam string vendorID
-- @tparam string deviceID
-- @treturn boolean requested value
function config.hasRTC(vendorID, deviceID)
    local result = config.getValue(vendorID, deviceID, config.HAS_RTC, 'Yes')
    return config.toBool(result)
end

function config.setSoftwareUpgradeStatus(vendorID, deviceID, softwareUpgradeStatus)
    config.storeValue(vendorID, deviceID, config.SOFTWARE_UPGRADE_STATUS, softwareUpgradeStatus)
end

function config.getSoftwareUpgradeRequest(vendorID, deviceID)
    return config.getValue(vendorID, deviceID, config.SOFTWARE_UPGRADE_REQUEST, '[]')
end

return config
