-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local tags = require 'utils/tags'

local deviceControl = {}

--- Stops the device thread for the specified vendor, device and sensor
-- @tparam string vendorID
-- @tparam string deviceID
-- @tparam string sensorID
function deviceControl.stopDeviceThread (vendorID, deviceID, sensorID)
    local TIMEOUT_TIMER = tags.getDeviceThreadTimer (vendorID, deviceID, sensorID)
    local DEVICE_THREAD = tags.getDeviceThread (vendorID, deviceID, sensorID)
    local THREAD_REPLY = tags.getDeviceThreadReply (vendorID, deviceID, sensorID)
    local keepWaiting = true

    eventHandler (THREAD_REPLY, function()
        keepWaiting = false
    end)
    eventHandler (TIMEOUT_TIMER, function ()
        WARN (string.format ("Waiting for thread '%s' to terminate timed out",
            DEVICE_THREAD))
        keepWaiting = false
    end)
    timerStart(TIMEOUT_TIMER, 5000)

    NOTIFYF ("Stopping thread %s", DEVICE_THREAD)
    eventPost(DEVICE_THREAD , "stop", true)

    while keepWaiting do
        local event, param = eventWait ()
        eventDispatch (event, param)
    end

    timerStop (TIMEOUT_TIMER)
    eventHandlerClear (TIMEOUT_TIMER)
    eventHandlerClear (THREAD_REPLY)
end

--- Starts the device thread for the specified vendor, device and sensor
-- @tparam string vendorID
-- @tparam string deviceID
-- @tparam string sensorID
-- @tparam function processEvent Callback function to be called for each event
function deviceControl.loopDeviceThread (vendorID, deviceID, sensorID,
        processEvent)
    local keepRunning = true
    local DEVICE_THREAD = tags.getDeviceThread (vendorID, deviceID, sensorID)
    local THREAD_REPLY = tags.getDeviceThreadReply (vendorID, deviceID, sensorID)

    local function stopThreadCallback (event, param)
        if param == "stop" then
            keepRunning = false
            eventPost (THREAD_REPLY, param)
        else
            WARNF ("%s, Ignored param '%s' on event", event, param)
        end
    end
    eventHandler (DEVICE_THREAD , stopThreadCallback)

    NOTIFYF ("device thread '%s' running", DEVICE_THREAD)

    while keepRunning do
        local event, param = eventWait ()
        if processEvent == nil or not processEvent (event, param) then
            eventDispatch (event, param)
        end
    end

    NOTIFYF ("device thread '%s' stopped", DEVICE_THREAD)
    eventHandlerClear (DEVICE_THREAD)
end

--- Starts the device thread for the specified vendor, device and sensor
-- @tparam string vendorID
-- @tparam string deviceID
-- @tparam string sensorID
-- @tparam function updateCallback Callback function to be called for each event
-- This function should return the actual sensor value
-- @tparam number updateIntervalMS Specifies internal sampling period in milliseconds
-- This is not the broker sample rate!
-- @tparam boolean async If true, samples will be sent whenever updateCallback returns
-- not nil value, ignoring the sample rate.
function deviceControl.loopSensorThread (vendorID, deviceID, sensorID,
        updateCallback, updateIntervalMS, async)
    updateIntervalMS = updateIntervalMS or 0
    local UPDATE_TIMER = tags.getDeviceThreadUpdateTimer (vendorID, deviceID,
        sensorID)

    local function processEvent (event, param)
        if event == UPDATE_TIMER then
            local value = updateCallback (event, param)
            if value or not async then
                sensorSet (sensorID, value)
            end
            if async then
                local asyncEvent = tags.getSensorAsyncUpdate(sensorID)
                tagSetString(asyncEvent, sensorID)
            end
            return true
        end
        return false
    end

    eventHandler (UPDATE_TIMER, function () end)
    -- post event to fetch initial data
    eventPost (UPDATE_TIMER)
    -- if updateCallback takes too long, event queue may
    -- overflow and crash
    timerPeriodic (UPDATE_TIMER, updateIntervalMS)

    deviceControl.loopDeviceThread (vendorID, deviceID, sensorID, processEvent)

    timerStop (UPDATE_TIMER)
    eventHandlerClear (UPDATE_TIMER)
end

--- Starts the device thread for the specified vendor, device and actuator
-- @tparam string vendorID
-- @tparam string deviceID
-- @tparam string sensorID
-- @tparam function updateCallback Callback function that receives the
-- actuator value and change actuator state for fitting this value
function deviceControl.loopActuatorThread (vendorID, deviceID, sensorID, updateCallback)

    local function processEvent (event, param)
        if event == sensorID then
            updateCallback (event, param)
            return true
        end
        return false
    end

    -- let thread listen to actuator events
    eventHandler (sensorID, function() end)

    deviceControl.loopDeviceThread (vendorID, deviceID, sensorID, processEvent)

    -- cleanup all events
    eventHandlerClear (sensorID)
end

return deviceControl
