-- CONFIDENTIAL AND PROPRIETARY INFORMATION

-- Copyright (c) 2013,2016 Emutex Ltd. All rights reserved.
-- This software and documentation contain confidential and
-- proprietary information that is the property of
-- Emutex Ltd. The software and documentation are
-- furnished under a license agreement and may be used
-- or copied only in accordance with the terms of the license
-- agreement. No part of the software and documentation
-- may be reproduced, transmitted, or translated, in any
-- form or by any means, electronic, mechanical, manual,
-- optical, or otherwise, without prior written permission
-- of Emutex Ltd., or as expressly provided by the license agreement.
-- Reverse engineering is prohibited, and reproduction,
-- disclosure or use without specific written authorization
-- of Emutex Ltd. is strictly forbidden.

local utils = {
    METADATA_VERSION = "1.0.0"
}

local boolConvertTable = {
    TRUE = true,
    FALSE = false
}

function utils.split(str, sep)
    sep = sep or "%s"
    local result = {}
    for i in string.gmatch(str, "[^" .. sep .. "]+") do
        table.insert(result, i)
    end
    return result
end

function utils.trim(s)
  return (s:gsub("^%s*(.-)%s*$", "%1"))
end

function utils.writeToFile(fileName, content)
    local f = io.open(fileName, "w")
    if not f then
        return false
    end
    f:write(content)
    f:close()
    return true
end

function utils.tobool(str)
    if not str or type(str) ~= 'string' then
        return nil
    end
    return boolConvertTable[str:upper()]
end

function utils.hexToBytes(str)
    if not str or str:len() % 2 ~= 0 then
        return nil
    end

    local bytes = {}
    local i = 1
    while i <= str:len() do
        local tmpstr = str:sub(i, i + 1)
        table.insert(bytes, string.char(tonumber(tmpstr, 16)))
        i = i + 2
    end

    return table.concat(bytes, "")
end

function utils.copyTable (src, dest)
    for key, value in pairs (src) do
        dest[key] = value
    end
end

function utils.copyShallow (orig)
    local copy = {}
    utils.copyTable (orig, copy)
    return copy
end

function utils.copyDeep (orig)
    local origType = type(orig)
    local copy
    if origType == 'table' then
        copy = {}
        for key, value in next, orig, nil do
            copy[utils.copyDeep(key)] = utils.copyDeep(value)
        end
        setmetatable(copy, utils.copyDeep(getmetatable(orig)))
    else
        copy = orig
    end
    return copy
end

function utils.mergeTables (table1, table2)
    local result = {}
    utils.copyTable (table1, result)
    utils.copyTable (table2, result)

    return result
end

function utils.toArray (t)
    if not t then
        return nil
    end

    local result = {}
    for key, value in pairs (t) do
        local numKey = tonumber (key)
        if not numKey then
            return nil
        end
        result[numKey] = value
    end

    return result
end

function utils.getScriptName ()
    return debug.getinfo(2, "S").short_src:gsub('%.lua', '')
end

function utils.tableLength (tbl)
    local count = 0
    for _ in pairs(tbl) do
        count = count + 1
    end
    return count
end

function utils.deepCompare (t1,t2,ignore_mt)
    local ty1 = type(t1)
    local ty2 = type(t2)
    if ty1 ~= ty2 then return false end
    -- non-table types can be directly compared
    if ty1 ~= 'table' and ty2 ~= 'table' then return t1 == t2 end
    -- as well as tables which have the metamethod __eq
    local mt = getmetatable(t1)
    if not ignore_mt and mt and mt.__eq then return t1 == t2 end
    for k1,v1 in pairs(t1) do
        local v2 = t2[k1]
        if v2 == nil or not utils.deepCompare(v1,v2) then return false end
    end
    for k2,v2 in pairs(t2) do
        local v1 = t1[k2]
        if v1 == nil or not utils.deepCompare(v1,v2) then return false end
    end
    return true
end

local function getScriptParams(value)
    local params = {}

    local regExp = '^ *(%w+) +"?([^"]+)"? *"?([^"]*)"?.*$'
    local ok, _, paramType, paramName, defValue = string.find(value, regExp)
    defValue = defValue or ""
    if paramType ~= "STR" then
        defValue = tonumber(defValue)
    end
    if ok then
        params = {
            type = paramType,
            name = paramName,
            defVal = defValue
        }
    end

    return params
end

function utils.getScriptMetadata(file)
    local metadata = {}
    local found = false
    local params = {}
    for line in io.lines(file) do
        if found then
            local ok, _, name, value = string.find(line, '^ *([^%]%[: %-\n]+): *([^\n]+)$')
            if ok then
                name = string.lower(name)
                local isParam, _, idx = string.find(name, 'param(%d+)')
                if isParam then
                    params[string.format("%s", idx)] = getScriptParams(value)
                else
                    metadata[name] = value
                end
            elseif string.find(line, '^ *%-%-%]%] *$') then
                break
            end
        elseif string.find(line, '^ *%-%-%[%[ *METADATA *$') then
            found = true
        end
    end

    if not found then
        WARNF("No metadata found at '%s'", file)
    end

    metadata.type = metadata.type or "Sensor"

    -- Ensures that all know metadata have a valid value
    metadata.devicename  = metadata.devicename           or file:gsub(".*/", ""):gsub("%.lua", "")
    metadata.description = metadata.description          or ""
    metadata.version     = metadata.version              or ""
    metadata.author      = metadata.author               or ""
    metadata.authorURI   = metadata.authorURI            or ""
    metadata.platform    = metadata.platform             or "Generic"
    metadata.datatype    = metadata.datatype             or "DEC"
    metadata.min         = tonumber(metadata.min)        or 0
    metadata.max         = tonumber(metadata.max)        or 100
    metadata.offset      = tonumber(metadata.offset)     or 0
    metadata.scale       = tonumber(metadata.scale)      or 1
    metadata.unit        = metadata.unit                 or ""
    metadata.samplerate  = tonumber(metadata.samplerate) or 10000
    metadata.params      = params

    return metadata
end

function utils.isIDReservedKeyword(ID)
    return string.sub(ID, 1, 1) == '_'
end

function utils.isIDReservedKeyword(ID)
    return string.sub(ID, 1, 1) == '_'
end

return utils
